""" Utilities for working with XML documents

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-13
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import lxml.etree
import re


def get_attributes_of_xpaths(filename, x_paths, attr='id'):
    """ Determine the number of objects that match each XPath in an XML file.

    Args:
        filename (:obj:`str`): path to XML file
        x_paths (:obj:`list` of `str`): XPaths
        attr (:obj:`str`, optional): attribute to get values of

    Returns:
        :obj:`dict` of :obj:`str` to :obj:`list` of :obj:`str`: dictionary that maps each XPath to the
            values of the attribute of the objects in the XML file that match the XPath
    """
    # read XML file
    etree = lxml.etree.parse(filename)

    # get namespaces
    root = etree.getroot()
    namespaces = root.nsmap
    if None in namespaces:
        namespaces.pop(None)
        match = re.match(r'^{(.*?)}(.*?)$', root.tag)
        if match:
            namespaces[match.group(2)] = match.group(1)

    # determine number of objects that match each XPath
    x_path_counts = {}
    for x_path in x_paths:
        try:
            objects = etree.xpath(x_path, namespaces=namespaces)

            x_path_counts[x_path] = [obj.attrib[attr] for obj in objects]
        except Exception:
            x_path_counts[x_path] = []

    # return number of objects that match each XPath
    return x_path_counts


def validate_xpaths_ref_to_unique_objects(filename, x_paths, attr='id'):
    """ Validate that each of a list of XPaths matches a single object in an XML file.

    Args:
        filename (:obj:`str`): path to XML file
        x_paths (:obj:`list` of `str`): XPaths
        attr (:obj:`str`, optional): attribute to get values of

    Returns:
        :obj:`dict` of :obj:`str` to :obj:`str`: dictionary that maps each XPath to the
            value of the attribute of the object in the XML file that matches the XPath

    Raises:
        :obj:`ValueError`: if one or more XPaths matches 0 or multiple objects
    """
    x_path_attr_vals = get_attributes_of_xpaths(filename, x_paths, attr=attr)

    errors = []

    no_matches = sorted([x_path for x_path, attr_vals in x_path_attr_vals.items() if len(attr_vals) == 0])
    if no_matches:
        errors.append('XPaths must reference unique objects. The following XPaths do not match any objects:\n  - {}'.format(
            '\n  - '.join(no_matches)))

    multiple_matches = sorted([x_path for x_path, attr_vals in x_path_attr_vals.items() if len(attr_vals) > 1])
    if multiple_matches:
        errors.append('XPaths must reference unique objects. The following XPaths match multiple objects:\n  - {}'.format(
            '\n  - '.join(multiple_matches)))

    if errors:
        raise ValueError('\n\n'.join(errors))

    return {key: vals[0] for key, vals in x_path_attr_vals.items()}



""" Utilities for executing tasks in SED-ML files in COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..archive.io import ArchiveWriter
from ..archive.utils import build_archive_from_paths
from ..config import get_config
from ..plot.data_model import PlotFormat  # noqa: F401
from ..report.data_model import DataGeneratorVariableResults, OutputResults, ReportFormat  # noqa: F401
from ..sedml.data_model import Task, DataGeneratorVariable  # noqa: F401
from .data_model import CombineArchiveContentFormatPattern
from .io import CombineArchiveReader
import biosimulators_utils.sedml.exec
import os
import re
import tempfile
import shutil
import types  # noqa: F401

__all__ = [
    'exec_sedml_docs_in_archive',
]


def exec_sedml_docs_in_archive(archive_filename, sed_task_executer, out_dir, apply_xml_model_changes=False,
                               report_formats=None, plot_formats=None,
                               bundle_outputs=None, keep_individual_outputs=None):
    """ Execute the SED-ML files in a COMBINE/OMEX archive (execute tasks and save outputs)

    Args:
        archive_filename (:obj:`str`): path to COMBINE/OMEX archive
        sed_task_executer (:obj:`types.FunctionType`): function to execute each SED task in each SED-ML file in the archive.
            The function must implement the following interface::

                def exec_task(task, variables):
                    ''' Execute a simulation and return its results

                    Args:
                       task (:obj:`Task`): task
                       variables (:obj:`list` of :obj:`DataGeneratorVariable`): variables that should be recorded

                    Returns:
                       :obj:`DataGeneratorVariableResults`: results of variables
                    '''
                    pass

        out_dir (:obj:`str`): path to store the outputs of the archive

            * CSV: directory in which to save outputs to files
              ``{ out_dir }/{ relative-path-to-SED-ML-file-within-archive }/{ report.id }.csv``
            * HDF5: directory in which to save a single HDF5 file (``{ out_dir }/reports.h5``),
              with reports at keys ``{ relative-path-to-SED-ML-file-within-archive }/{ report.id }`` within the HDF5 file

        apply_xml_model_changes (:obj:`bool`): if :obj:`True`, apply any model changes specified in the SED-ML files before
            calling :obj:`task_executer`.
        report_formats (:obj:`list` of :obj:`ReportFormat`, optional): report format (e.g., csv or h5)
        plot_formats (:obj:`list` of :obj:`PlotFormat`, optional): report format (e.g., pdf)
        bundle_outputs (:obj:`bool`, optional): if :obj:`True`, bundle outputs into archives for reports and plots
        keep_individual_outputs (:obj:`bool`, optional): if :obj:`True`, keep individual output files
    """
    config = get_config()

    # process arguments
    if report_formats is None:
        report_formats = [ReportFormat(format_value) for format_value in config.REPORT_FORMATS]

    if plot_formats is None:
        plot_formats = [PlotFormat(format_value) for format_value in config.PLOT_FORMATS]

    if bundle_outputs is None:
        bundle_outputs = config.BUNDLE_OUTPUTS

    if keep_individual_outputs is None:
        keep_individual_outputs = config.KEEP_INDIVIDUAL_OUTPUTS

    # create temporary directory to unpack archive
    archive_tmp_dir = tempfile.mkdtemp()

    # unpack archive and read metadata
    archive = CombineArchiveReader.run(archive_filename, archive_tmp_dir)

    # determine files to execute
    master_content = archive.get_master_content()
    exec_content = [master_content] if master_content else archive.contents

    # execute SED-ML files: execute tasks and save outputs
    tmp_out_dir = tempfile.mkdtemp()
    for content in exec_content:
        if re.match(CombineArchiveContentFormatPattern.SED_ML.value, content.format):
            if os.path.isabs(content.location):
                raise ValueError('Content locations must be relative')
            content_filename = os.path.join(archive_tmp_dir, content.location)
            working_dir = os.path.dirname(content_filename)
            biosimulators_utils.sedml.exec.exec_doc(content_filename,
                                                    working_dir,
                                                    sed_task_executer,
                                                    tmp_out_dir,
                                                    os.path.relpath(content_filename, archive_tmp_dir),
                                                    apply_xml_model_changes=apply_xml_model_changes,
                                                    report_formats=report_formats,
                                                    plot_formats=plot_formats)

    # arrange outputs
    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)

    # move HDF5 file to desired location
    tmp_hdf5_path = os.path.join(tmp_out_dir, config.H5_REPORTS_PATH)
    out_hdf5_path = os.path.join(out_dir, config.H5_REPORTS_PATH)
    if os.path.isfile(tmp_hdf5_path):
        shutil.move(tmp_hdf5_path, out_hdf5_path)

    if bundle_outputs:
        # bundle CSV files of reports into zip archive
        archive_paths = [os.path.join(tmp_out_dir, '**', '*.' + format.value) for format in report_formats if format != ReportFormat.h5]
        archive = build_archive_from_paths(archive_paths, tmp_out_dir)
        if archive.files:
            ArchiveWriter().run(archive, os.path.join(out_dir, config.REPORTS_PATH))

        # bundle PDF files of plots into zip archive
        archive_paths = [os.path.join(tmp_out_dir, '**', '*.' + format.value) for format in plot_formats]
        archive = build_archive_from_paths(archive_paths, tmp_out_dir)
        if archive.files:
            ArchiveWriter().run(archive, os.path.join(out_dir, config.PLOTS_PATH))

    # cleanup temporary files
    if keep_individual_outputs:
        if os.path.isdir(out_dir):
            for file_or_dir in os.listdir(tmp_out_dir):
                shutil.move(os.path.join(tmp_out_dir, file_or_dir), out_dir)
        else:  # pragma: no cover # unreachable because the above code creates :obj:`out_dir` if it doesn't exist
            shutil.move(tmp_out_dir, out_dir)

    else:
        shutil.rmtree(tmp_out_dir)

    shutil.rmtree(archive_tmp_dir)



""" Utilities for creating BioSimulators-compliant command-line interfaces for biosimulation tools.

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-04
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import cement
import sys
import types  # noqa: F401

__all__ = [
    'build_cli',
]


def build_cli(cli_name=None, cli_version=None,
              simulator_name=None, simulator_version=None, simulator_url=None,
              combine_archive_executer=None):
    """ Create a BioSimulators-compliant command-line application for a biosimulation tool.

    The command-line application will have two inputs

    * A path to a COMBINE/OMEX archive that describes one or more simulations or one or more models
    * A path to a directory to store the ouputs of the execution of the simulations defined in the archive

    The command-line application will also support two additional commands

    * A command for printing help information about the command-line application (`-h`, `--help`)
    * A command for printing version information about the command-line application (`-v`, `--version`)

    Args:
        cli_name (:obj:`str`): name of the command-line program (e.g., `copasi`)
        cli_version (:obj:`str`, optional): version of the command-line application
        simulator_name (:obj:`str`): name of the simulator (e.g., `COPASI`)
        simulator_version (:obj:`str`, optional): version of the simulator
        simulator_url (:obj:`str`, optional): URL for information about the simulator
        combine_archive_executer (:obj:`types.FunctionType`): a function which has two positional arguments
            * The path to the COMBINE/OMEX archive
            * The path to the directory to save the outputs of the simulations defined in the archive

    Returns:
        :obj:`cement.App`: command-line application

    Raises:
        :obj:`SystemExit`: if the execution of the COMBINE/OMEX archive fails
    """
    python_version = '.'.join(str(i) for i in sys.version_info[0:3])

    if not cli_name:
        raise ValueError('CLI name must be defined')
    if not simulator_name:
        raise ValueError('Simulator name must be defined')
    if not combine_archive_executer:
        raise ValueError('COMBINE/OMEX archive execution method must be provided')

    versions = []
    if simulator_version:
        versions.append(simulator_name + ': ' + simulator_version)
    if cli_version:
        versions.append('CLI: ' + cli_version)
    versions.append('Python: ' + python_version)
    version = ', '.join(versions)

    class BaseController(cement.Controller):
        """ Base controller for command line application """

        class Meta:
            label = 'base'
            description = "BioSimulators-compliant command-line interface to the {} simulation program{}.".format(
                simulator_name,
                '<{}>'.format(simulator_url) if simulator_url else '')
            help = cli_name
            arguments = [
                (
                    ['-i', '--archive'],
                    dict(
                        type=str,
                        required=True,
                        help='Path to COMBINE/OMEX file which contains one or more SED-ML-encoded simulation experiments',
                    ),
                ),
                (
                    ['-o', '--out-dir'],
                    dict(
                        type=str,
                        default='.',
                        help='Directory to save outputs',
                    ),
                ),
                (
                    ['-v', '--version'],
                    dict(
                        action='version',
                        version=version,
                    ),
                ),
            ]

        @cement.ex(hide=True)
        def _default(self):
            args = self.app.pargs
            try:
                combine_archive_executer(args.archive, args.out_dir)
            except Exception as exception:
                raise SystemExit(str(exception)) from exception

    class App(cement.App):
        """ Command line application """
        class Meta:
            label = 'cli'
            base_controller = 'base'
            handlers = [
                BaseController,
            ]

    return App



""" Data model for submitting simulators to the BioSimulators registry

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import enum


class IssueLabel(str, enum.Enum):
    validated = 'Validated'
    invalid = 'Invalid'
    approved = 'Approved'
    action_error = 'Action error'


class SimulatorSubmission(object):
    """ Submission of a simulator to the BioSimulators registry

    Attributes:
        id (:obj:`str`): id of simulator (e.g., `tellurium` or `vcell`)
        version (:obj:`str`): version of simulator (e.g., `2.1.6`)
        specifications_url (:obj:`str`): URL for the specifications of the version of the simulator
            (e.g., `https://raw.githubusercontent.com/biosimulators/Biosimulators_tellurium/2.1.6/biosimulators.json`)
        specifications_patch (:obj:`dict`): superseding specifications to those at :obj:`specifications_url`
        validate_image (:obj:`bool`): if :obj:`True`, validate Docker image
        commit_simulator (:obj:`bool`): if :obj:`True`, commit simulator to database
    """

    def __init__(self, id=None, version=None, specifications_url=None,
                 specifications_patch=None,
                 validate_image=False, commit_simulator=False,
                 validated=False, approved=False, committed=False):
        """
        Args:
            id (:obj:`str`, optional): id of simulator
            version (:obj:`str`, optional): version of simulator
            specifications_url (:obj:`str`, optional): URL for the specifications of the version of the simulator
            specifications_patch (:obj:`dict`, optional): superseding specifications to those at :obj:`specifications_url`
            validate_image (:obj:`bool`, optional): if :obj:`True`, validate Docker image
            commit_simulator (:obj:`bool`, optional): if :obj:`True`, commit simulator to database
        """
        self.id = id
        self.version = version
        self.specifications_url = specifications_url
        self.specifications_patch = specifications_patch or {}
        self.validate_image = validate_image
        self.commit_simulator = commit_simulator

    def to_tuple(self):
        """ Tuple representation of a person

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation of a person
        """
        return (self.id, self.version, self.specifications_url, self.specifications_patch, self.validate_image, self.commit_simulator)

    def is_equal(self, other):
        """ Determine if two submissions are equal

        Args:
            other (:obj:`SimulatorSubmission`): another submission

        Returns:
            :obj:`bool`: :obj:`True`, if two submissions are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.version == other.version \
            and self.specifications_url == other.specifications_url \
            and self.specifications_patch == other.specifications_patch \
            and self.validate_image == other.validate_image \
            and self.commit_simulator == other.commit_simulator



""" Methods for processing submissions to the BioSimulators registry

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import SimulatorSubmission
import io
import yamldown

__all__ = ['get_simulator_submission_from_gh_issue_body']


def get_simulator_submission_from_gh_issue_body(body):
    """ Get a simulator submission from the YAML-structured data in an issue

    Args:
       body (:obj:`str`): body of a GitHub issue for the submission of a simulator

    Returns:
        :obj:`SimulatorSubmission`: simulator submission
    """
    body_stream = io.StringIO(body.replace('\r', ''))
    data, _ = yamldown.load(body_stream)
    return get_simulator_submission_from_gh_issue_body_data(data)


def get_simulator_submission_from_gh_issue_body_data(body_data):
    """ Get a simulator submission from the YAML-structured data in an issue

    Args:
       body (:obj:`object`): data from the body of a GitHub issue for the submission of a simulator

    Returns:
        :obj:`SimulatorSubmission`: simulator submission

    Raises:
        :obj:`ValueError`: if submission is invalid (e.g., missing or extra keys)
    """
    id = body_data.get('id', None) or None
    version = body_data.get('version', None) or None
    specifications_url = body_data.get('specificationsUrl', None) or None
    specifications_patch = body_data.get('specificationsPatch', {}) or {}
    validate_image = body_data.get('validateImage', False)
    commit_simulator = body_data.get('commitSimulator', False)

    # validate properties of submission
    errors = []

    if not id:
        errors.append("Simulator submissions must provide the id for the simulator.")

    if not version:
        errors.append("Simulator submissions must provide the version for the simulator.")

    if not specifications_url:
        errors.append("Simulator submissions must provide a URL where the specifications of the simulator can be downloaded.")

    allowed_keys = set(['id', 'version', 'specificationsUrl', 'specificationsPatch', 'validateImage', 'commitSimulator'])
    extra_keys = set(body_data.keys()).difference(allowed_keys)
    if extra_keys:
        errors.append(('Simulator submissions should only use the keys '
                       '"id", "version", "specificationsUrl", "specificationsPatch", validateImage", and "commitSimulator". '
                       'The following keys are not valid:\n  - {}'.format('\n  - '.join(sorted(extra_keys)))))

    if errors:
        raise ValueError("\n\n".join(errors))

    # return submission
    return SimulatorSubmission(id=id, version=version, specifications_url=specifications_url,
                               specifications_patch=specifications_patch,
                               validate_image=validate_image, commit_simulator=commit_simulator)



""" Utilities for working with Docker images

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import docker
import json
import os
import subprocess

__all__ = [
    'login_to_docker_registry',
    'pull_docker_image',
    'tag_and_push_docker_image',
    'convert_docker_image_to_singularity',
]


def login_to_docker_registry(registry, username, password):
    """ Login to a Docker registry

    Args:
        registry (:obj:`str`): registry (e.g., ghcr.io)
        username (:obj:`str`): user name
        password (:obj:`str`): password

    Returns:
        :obj:`docker.client.DockerClient`: Docker client
    """
    docker_client = docker.from_env()
    docker_client.login(registry=registry, username=username, password=password)
    return docker_client


def pull_docker_image(docker_client, url):
    """ Pull Docker image

    Args:
        docker_client (:obj:`docker.client.DockerClient`): Docker client
        url (:obj:`str`): URL for Docker image

    Returns:
        :obj:`docker.models.images.Image`: Docker image
    """
    try:
        return docker_client.images.pull(url)
    except docker.errors.NotFound:
        raise docker.errors.NotFound('{} is not a tag or URL for a Docker image.'.format(url))
    except Exception as error:
        raise Exception('Image {} could not be pulled.\n\n  {}'.format(
            url, str(error).replace('\n', '\n  ')))


def tag_and_push_docker_image(docker_client, image, tag):
    """ Tag and push Docker image

    Args:
        docker_client (:obj:`docker.client.DockerClient`): Docker client
        image (:obj:`docker.models.images.Image`): Docker image
        tag (:obj:`str`): tag
    """
    assert image.tag(tag)
    response = docker_client.images.push(tag)
    response = json.loads(response.rstrip().split('\n')[-1])
    if 'error' in response:
        raise Exception('Unable to push image to {}\n  {}'.format(
            tag, response.get('errorDetail', {}).get('message', response['error'])))


def convert_docker_image_to_singularity(docker_image, singularity_filename=None):
    """ Convert a Docker image to a Singularity image

    Args:
        docker_image (:obj:`str`)
        singularity_filename (:obj:`str`, optional): file name for saving Singularity image

    Returns:
        :obj:`str`: path where Singularity image was saved
    """
    cmd = ['singularity', 'pull']

    if not singularity_filename:
        singularity_filename = os.path.join(
            os.path.expanduser('~'), '.biosimulators-utils', 'singularity',
            docker_image.replace('/', '_').replace(':', '_') + '.sif')

    if not os.path.isdir(os.path.dirname(singularity_filename)):
        os.makedirs(os.path.dirname(singularity_filename))

    cmd.append(singularity_filename)

    cmd.append('docker://' + docker_image)

    if not os.path.isfile(singularity_filename):
        subprocess.check_call(cmd)

    return singularity_filename



""" Utilities for creating archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import Archive, ArchiveFile
import glob
import os

__all__ = ['build_archive_from_paths']


def build_archive_from_paths(path_patterns, rel_path=None, recursive=True):
    """ Build an archive from a list of glob path patterns

    Args:
        path_patterns (:obj:`list` of :obj:`str`): glob path patterns for files to bundle into an archive
        rel_path (:obj:`str`, optional): if provided, set the archive file names to their path relative to this path
        recursive (:obj:`bool`, optional): if :obj:`True`, match the path patterns recursively

    Returns:
        :obj:`Archive`: archive
    """
    archive = Archive()
    for path_pattern in path_patterns:
        for local_path in glob.glob(path_pattern, recursive=recursive):
            if os.path.isfile(local_path):
                if rel_path:
                    archive_path = os.path.relpath(local_path, rel_path)
                else:
                    archive_path = local_path

                archive.files.append(ArchiveFile(
                    local_path=local_path,
                    archive_path=archive_path,
                ))
    return archive



""" Data model for simulators

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import enum

__all__ = ['SoftwareInterface']


class SoftwareInterface(str, enum.Enum):
    """ Software interface """
    library = 'library'
    command_line = 'command-line application'
    desktop_application = 'desktop application'
    mobile_application = 'mobile application'
    web_service = 'web service'
    web_application = 'web application'
    biosimulators_docker_image = 'BioSimulators Docker image'



""" Core data model

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import enum

__all__ = [
    'Person',
    'Identifier',
    'OntologyTerm',
    'ValueType',
]


class Person(object):
    """ A person

    Attributes:
        given_name (:obj:`str`): given/first name
        other_name (:obj:`str`): other/middle name
        family_name (:obj:`str`): family/last name
    """

    def __init__(self, given_name=None, other_name=None, family_name=None):
        """
        Args:
            given_name (:obj:`str`, optional): given/first name
            other_name (:obj:`str`, optional): other/middle name
            family_name (:obj:`str`, optional): family/last name
        """
        self.given_name = given_name
        self.other_name = other_name
        self.family_name = family_name

    def to_tuple(self):
        """ Tuple representation of a person

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation of a person
        """
        return (self.family_name, self.other_name, self.given_name)

    def is_equal(self, other):
        """ Determine if two authors are equal

        Args:
            other (:obj:`Person`): another author

        Returns:
            :obj:`bool`: :obj:`True`, if two authors are equal
        """
        return self.__class__ == other.__class__ \
            and self.given_name == other.given_name \
            and self.other_name == other.other_name \
            and self.family_name == other.family_name


class Identifier(object):
    """ An identifier

    Attributes:
        namespace (:obj:`str`): namespace
        id (:obj:`str`): id
        url (:obj:`str`): URL
    """

    def __init__(self, namespace=None, id=None, url=None):
        """
        Args:
            namespace (:obj:`str`, optional): namespace
            id (:obj:`str`, optional): id
            url (:obj:`str`, optional): URL
        """
        self.namespace = namespace
        self.id = id
        self.url = url

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.namespace, self.id, self.url)

    def is_equal(self, other):
        """ Determine if identifiers are equal

        Args:
            other (:obj:`Identifier`): another identifier

        Returns:
            :obj:`bool`: :obj:`True`, if two identifiers are equal
        """
        return self.__class__ == other.__class__ \
            and self.namespace == other.namespace \
            and self.id == other.id \
            and self.url == other.url


class OntologyTerm(Identifier):
    """ Term in an ontology """
    pass


class ValueType(str, enum.Enum):
    """ A type of value (e.g., of a model attribute change or algorithm parameter change) """
    boolean = 'boolean'
    integer = 'integer'
    float = 'float'
    string = 'string'
    kisao_id = 'kisaoId'
    list = 'list'
    object = 'object'
    any = 'any'



""" Utilities for executing command-line interfaces to simulators and
containers that contain command-line interfaces to simulators

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

try:
    import docker
except ModuleNotFoundError:
    docker = None
import os
import shutil
import subprocess
import tempfile

__all__ = [
    'exec_sedml_docs_in_archive_with_simulator_cli',
    'exec_sedml_docs_in_archive_with_containerized_simulator',
]


def exec_sedml_docs_in_archive_with_simulator_cli(archive_filename, out_dir, simulator_command):
    """ Use a command-line interface to a simulation tool to execute the tasks specified in archive
    COMBINE/OMEX archive and generate the reports specified in the archive

    Args:
        archive_filename (:obj:`str`): path to a COMBINE/OMEX archive
        out_dir (:obj:`str`): directory where outputs should be saved
        simulator_command (:obj:`str`): system command for the simulator to execute

    Raises:
        :obj:`RuntimeError`: if the execution failed
    """
    try:
        subprocess.check_call([simulator_command] + build_cli_args(archive_filename, out_dir))

    except FileNotFoundError:
        raise RuntimeError("The command '{}' could not be found".format(simulator_command))

    except subprocess.CalledProcessError as exception:
        raise RuntimeError("The command '{}' could not execute the archive:\n\n  {}".format(
            simulator_command, exception.stderr.replace('\n', '\n  ')))

    except Exception as exception:
        raise RuntimeError("The command '{}' could not execute the archive:\n\n  {}".format(
            simulator_command, str(exception).replace('\n', '\n  ')))


def exec_sedml_docs_in_archive_with_containerized_simulator(
        archive_filename, out_dir, docker_image,
        docker_image_temp_dir='/tmp', docker_image_path_sep='/',
        environment=None, pull_docker_image=True,
        user_to_exec_within_container='_CURRENT_USER_', allocate_tty=True, remove_docker_container=True):
    """ Use a containerized simulator tool to execute the tasks specified in a
    COMBINE/OMEX archive and generate the reports specified in the archive

    Args:
        archive_filename (:obj:`str`): path to a COMBINE/OMEX archive
        out_dir (:obj:`str`): directory where outputs should be saved
        docker_image (:obj:`str`): tag (e.g., ``biosimulators/tellurium``) or
            URL (``ghcr.io/biosimulators/tellurium``) for a Docker image of a simulator
        docker_image_temp_dir (:obj:`str`, optional): Path to the temporary directory within the Docker image
            (e.g., ``/tmp`` for Linux images, ``C:\\Users\\{ user }\\AppData\\Local\\Temp`` for Windows images).
            The path can either be an absolute path or a path relative to the working directory of the image.
        docker_image_path_sep (:obj:`str`, optional): Path separator for the image (e.g., ``/`` for
            Linux, Mac OS, Unix images; ``\\`` for Windows images)
        environment (:obj:`dict`, optional): environment variables for executing the Docker image
        pull_docker_image (:obj:`bool`, optional): if :obj:`True`, pull the Docker image (if the image isn't
            available locally, this will cause the image to be downloaded; this will cause the image to be updated)
        user_to_exec_within_container (:obj:`str`, optional): username or user id to execute commands within the Docker container

            * Use ``_CURRENT_USER_`` to indicate that the Docker container should execute commands as the current user (``os.getuid()``)
            * Use the format ``<name|uid>[:<group|gid>]`` to indicate any other user/group that the Docker container should use to
              execute commands

        allocate_tty (:obj:`bool`, optional): if :obj:`True`, allocate a pseudo-TTY
        remove_docker_container (:obj:`bool`, optional): if :obj:`True`, automatically remove the container when it exits

    Raises:
        :obj:`RuntimeError`: if the execution failed
    """
    if not docker:
        raise ModuleNotFoundError("No module named 'docker'. Docker and the Python Docker package must be installed.")

    # pull image
    get_simulator_docker_image(docker_image, pull=pull_docker_image)

    # run image
    in_dir = tempfile.mkdtemp()
    if os.path.isfile(archive_filename):
        shutil.copyfile(archive_filename, os.path.join(in_dir, os.path.basename(archive_filename)))

    if not os.path.isdir(out_dir):
        os.makedirs(out_dir)
    image_in_dir = docker_image_path_sep.join((docker_image_temp_dir, 'in'))
    image_out_dir = docker_image_path_sep.join((docker_image_temp_dir, 'out'))

    # setup Docker run arguments
    args = ['docker', 'run']

    args.extend(['--mount', 'type=bind,source={},target={},readonly'.format(in_dir, image_in_dir)])
    args.extend(['--mount', 'type=bind,source={},target={}'.format(os.path.abspath(out_dir), image_out_dir)])

    if environment:
        for key, val in environment.items():
            args.extend(['--env', '{}={}'.format(key, val)])

    if user_to_exec_within_container:
        if user_to_exec_within_container == '_CURRENT_USER_':
            if os.name == 'posix':
                args.extend(['--user', str(os.getuid())])
            else:
                raise NotImplementedError('The current user id can only be retrieved for POSIX OSes')
        else:
            args.extend(['--user', user_to_exec_within_container])

    if allocate_tty:
        args.append('--tty')

    if remove_docker_container:
        args.append('--rm')

    args.append(docker_image)
    args.extend(build_cli_args(
        docker_image_path_sep.join((image_in_dir, os.path.basename(archive_filename))),
        image_out_dir,
    ))

    # run image
    try:
        subprocess.check_call(args)

    except FileNotFoundError:
        raise RuntimeError("Docker could not be found")

    except subprocess.CalledProcessError as exception:
        raise RuntimeError("The image '{}' could not execute the archive:\n\n  {}".format(
            docker_image, (exception.stderr or 'Unknown error').replace('\n', '\n  ')))

    except Exception as exception:
        raise RuntimeError("The image '{}' could not execute the archive:\n\n  {}".format(
            docker_image, str(exception).replace('\n', '\n  ')))

    finally:
        shutil.rmtree(in_dir)


def build_cli_args(archive_filename, out_dir):
    """ Build the arguments to use a command-line interface to a simulator to execute a COMBINE/OMEX archive

    Args:
        archive_filename (:obj:`str`): path to a COMBINE/OMEX archive
        out_dir (:obj:`str`): directory where outputs should be saved

    Returns:
        :obj:`list` of :obj:`str`: command-line arguments to execute a COMBINE/OMEX archive
    """
    return [
        '-i', archive_filename,
        '-o', out_dir,
    ]


def get_simulator_docker_image(tag, pull=True):
    """ Get a Docker image for a simulator

    Args:
        tag (:obj:`str`): tag (e.g., ``biosimulators/tellurium``) or
            URL (``ghcr.io/biosimulators/tellurium`) for a Docker image of a simulator
    Returns:
        :obj:`docker.models.images.Image`: Docker image
    """
    docker_client = docker.from_env()

    try:
        image = docker_client.images.get(tag)
        if pull:
            try:
                image = docker_client.images.pull(tag)
            except Exception:  # pragma: no cover
                pass

    except Exception:
        if pull:
            try:
                image = docker_client.images.pull(tag)
            except Exception:
                raise docker.errors.ImageNotFound("Image '{}' for simulator could not be pulled".format(tag))
        else:
            raise docker.errors.ImageNotFound("Image '{}' for simulator is not available locally".format(tag))

    return image



""" Methods for querying the BioSimulators registry

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config
import requests


__all__ = ['get_simulator_version_specs']


def get_simulator_version_specs(id):
    """ Get the specifications of the versions of a simulator in the BioSimulators registry

    Args:
        id (:obj:`str`): simulator id

    Returns:
        :obj:`list` of :obj:`dict`: specifications of the registered versions of the simulator
    """

    endpoint = get_config().BIOSIMULATORS_API_ENDPOINT
    response = requests.get('{}simulators/{}'.format(endpoint, id))
    try:
        response.raise_for_status()
        version_specs = response.json()
    except requests.exceptions.HTTPError:
        if response.status_code != 404:
            raise
        version_specs = []
    return version_specs



""" Data model for SED

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..biosimulations.data_model import Metadata  # noqa: F401
from ..utils.core import are_lists_equal, none_sorted
import abc
import enum
import numpy  # noqa: F401


__all__ = [
    'ModelLanguage',
    'DataGeneratorVariableSymbol',
    'SedDocument',
    'Simulation',
    'SteadyStateSimulation',
    'OneStepSimulation',
    'UniformTimeCourseSimulation',
    'Algorithm',
    'AlgorithmParameterChange',
    'Model',
    'ModelChange',
    'ModelAttributeChange',
    'AbstractTask',
    'Task',
    'DataGenerator',
    'DataGeneratorVariable',
    'DataGeneratorParameter',
    'Output',
    'Report',
    'DataSet',
    'Plot2D',
    'Plot3D',
    'AxisScale',
    'Curve',
    'Surface',
]


class ModelLanguage(str, enum.Enum):
    """ Model language """
    BGNL = 'urn:sedml:language:bgnl'
    CellML = 'urn:sedml:language:cellml'
    NeuroML = 'urn:sedml:language:neuroml'
    SBML = 'urn:sedml:language:sbml'
    VCML = 'urn:sedml:language:vcml'


class DataGeneratorVariableSymbol(str, enum.Enum):
    """ Variable sumbol """
    time = 'urn:sedml:symbol:time'


class SedDocument(object):
    """ A SED-ML document

    Attributes:
        level (:obj:`int`): level
        version (:obj:`int`): version
        models (:obj:`list` of :obj:`Model`): models
        simulations (:obj:`list` of :obj:`Simulation`): simulations
        tasks (:obj:`list` of :obj:`AbstractTask`): tasks
        data_generators (:obj:`list` of :obj:`DataGenerator`): data generators
        outputs (:obj:`list` of :obj:`Output`): outputs
        metadata (:obj:`Metadata`): metadata
    """

    def __init__(self, level=1, version=3, models=None, simulations=None, tasks=None, data_generators=None, outputs=None, metadata=None):
        """
        Args:
            level (:obj:`int`, optional): level
            version (:obj:`int`, optional): version
            models (:obj:`list` of :obj:`Model`, optional): models
            simulations (:obj:`list` of :obj:`Simulation`, optional): simulations
            tasks (:obj:`list` of :obj:`AbstractTask`, optional): tasks
            data_generators (:obj:`list` of :obj:`DataGenerator`, optional): data generators
            outputs (:obj:`list` of :obj:`Output`, optional): outputs
            metadata (:obj:`Metadata`, optional): metadata
        """
        self.level = level
        self.version = version
        self.models = models or []
        self.simulations = simulations or []
        self.tasks = tasks or []
        self.data_generators = data_generators or []
        self.outputs = outputs or []
        self.metadata = metadata

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (
            self.level,
            self.version,
            tuple(none_sorted(model.to_tuple() for model in self.models)),
            tuple(none_sorted(simulation.to_tuple() for simulation in self.simulations)),
            tuple(none_sorted(task.to_tuple() for task in self.tasks)),
            tuple(none_sorted(data_generator.to_tuple() for data_generator in self.data_generators)),
            tuple(none_sorted(output.to_tuple() for output in self.outputs)),
            self.metadata.to_tuple() if self.metadata else None,
        )

    def is_equal(self, other):
        """ Determine if SED-ML documents are equal

        Args:
            other (:obj:`Simulation`): another SED-ML document

        Returns:
            :obj:`bool`: :obj:`True`, if two SED-ML documents are equal
        """
        return self.__class__ == other.__class__ \
            and self.level == other.level \
            and self.version == other.version \
            and are_lists_equal(self.models, other.models) \
            and are_lists_equal(self.simulations, other.simulations) \
            and are_lists_equal(self.tasks, other.tasks) \
            and are_lists_equal(self.data_generators, other.data_generators) \
            and are_lists_equal(self.outputs, other.outputs) \
            and ((self.metadata is None and self.metadata == other.metadata)
                 or (self.metadata is not None and self.metadata.is_equal(other.metadata)))


class Simulation(abc.ABC):
    """ A simulation

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        algorithm (:obj:`Algorithm`): algorithm
    """

    def __init__(self, id=None, name=None, algorithm=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            algorithm (:obj:`Algorithm`, optional): algorithm
        """
        self.id = id
        self.name = name
        self.algorithm = algorithm

    @abc.abstractmethod
    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        pass  # pragma: no cover

    def is_equal(self, other):
        """ Determine if simulations are equal

        Args:
            other (:obj:`Simulation`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two simulations are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and ((self.algorithm is None and self.algorithm == other.algorithm)
                 or (self.algorithm is not None and self.algorithm.is_equal(other.algorithm)))


class SteadyStateSimulation(Simulation):
    """ A steady-state simulation

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        algorithm (:obj:`Algorithm`): algorithm
    """

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name, self.algorithm.to_tuple() if self.algorithm else None)


class OneStepSimulation(Simulation):
    """ A single simulation step

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        algorithm (:obj:`Algorithm`): algorithm
        step (:obj:`float`): step
    """

    def __init__(self, id=None, name=None, algorithm=None, step=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            algorithm (:obj:`Algorithm`, optional): algorithm
            step (:obj:`float`, optional): step
        """
        super(OneStepSimulation, self).__init__(id=id, name=name, algorithm=algorithm)
        self.step = step

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name, self.algorithm.to_tuple() if self.algorithm else None, self.step)

    def is_equal(self, other):
        """ Determine if simulations are equal

        Args:
            other (:obj:`OneStepSimulation`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two simulations are equal
        """
        return super(OneStepSimulation, self).is_equal(other) \
            and self.step == other.step


class UniformTimeCourseSimulation(Simulation):
    """ A uniform time course simulation

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        algorithm (:obj:`Algorithm`): algorithm
        initial_time (:obj:`float`): initial time
        output_start_time (:obj:`float`): output start time
        output_end_time (:obj:`float`): output end time
        number_of_points (:obj:`int`): number of time points
    """

    def __init__(self, id=None, name=None, algorithm=None,
                 initial_time=None, output_start_time=None, output_end_time=None, number_of_points=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            algorithm (:obj:`Algorithm`, optional): algorithm
            initial_time (:obj:`float`, optional): initial time
            output_start_time (:obj:`float`, optional): output start time
            output_end_time (:obj:`float`, optional): output end time
            number_of_points (:obj:`int`, optional): number of time points
        """
        super(UniformTimeCourseSimulation, self).__init__(id=id, name=name, algorithm=algorithm)
        self.initial_time = initial_time
        self.output_start_time = output_start_time
        self.output_end_time = output_end_time
        self.number_of_points = number_of_points

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name, self.algorithm.to_tuple() if self.algorithm else None,
                self.initial_time, self.output_start_time, self.output_end_time, self.number_of_points)

    def is_equal(self, other):
        """ Determine if simulations are equal

        Args:
            other (:obj:`UniformTimeCourseSimulation`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two simulations are equal
        """
        return super(UniformTimeCourseSimulation, self).is_equal(other) \
            and self.initial_time == other.initial_time \
            and self.output_start_time == other.output_start_time \
            and self.output_end_time == other.output_end_time \
            and self.number_of_points == other.number_of_points


class Algorithm(object):
    """ A simulation algorithm

    Attributes:
        kisao_id (:obj:`str`): KiSAO id (e.g., `KISAO_0000029`)
        changes (:obj:`list` of :obj:`AlgorithmParameterChange`): parameter changes
    """

    def __init__(self, kisao_id=None, changes=None):
        """
        Args:
            kisao_id (:obj:`str`, optional): KiSAO id (e.g., `KISAO_0000029`)
            changes (:obj:`list` of :obj:`AlgorithmParameterChange`, optional): parameter changes
        """
        self.kisao_id = kisao_id
        self.changes = changes or []

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.kisao_id,
                tuple(none_sorted(change.to_tuple() for change in self.changes)))

    def is_equal(self, other):
        """ Determine if algorithms are equal

        Args:
            other (:obj:`Algorithm`): another algorithm

        Returns:
            :obj:`bool`: :obj:`True`, if two algorithms are equal
        """
        return self.__class__ == other.__class__ \
            and self.kisao_id == other.kisao_id \
            and are_lists_equal(self.changes, other.changes)


class AlgorithmParameterChange(object):
    """ A changed parameter of an algorithm

    Attributes:
        kisao_id (:obj:`str`): KiSAO id (e.g., `KISAO_0000029`)
        new_value (:obj:`str`): new value
    """

    def __init__(self, kisao_id=None, new_value=None):
        """
        Args:
            kisao_id (:obj:`str`, optional): KiSAO id (e.g., `KISAO_0000029`)
            new_value (:obj:`str`, optional): new value
        """
        self.kisao_id = kisao_id
        self.new_value = new_value

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.kisao_id, self.new_value)

    def is_equal(self, other):
        """ Determine if parameter changes are equal

        Args:
            other (:obj:`AlgorithmParameterChange`): another parameter change

        Returns:
            :obj:`bool`: :obj:`True`, if two parameter changes are equal
        """
        return self.__class__ == other.__class__ \
            and self.kisao_id == other.kisao_id \
            and self.new_value == other.new_value


class Model(object):
    """ A model

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        source (:obj:`str`): path to the model file
        language (:obj:`str`): URN of the format of the model
        changes (:obj:`list` of :obj:`ModelChange`): model changes
    """

    def __init__(self, id=None, name=None, source=None, language=None, changes=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            source (:obj:`str`), optional: path to the model file
            language (:obj:`str`, optional): URN of the format of the model
            changes (:obj:`list` of :obj:`ModelChange`, optional): model changes
        """
        self.id = id
        self.name = name
        self.source = source
        self.language = language
        self.changes = changes or []

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name, self.source, self.language,
                tuple(none_sorted(change.to_tuple() for change in self.changes)))

    def is_equal(self, other):
        """ Determine if models are equal

        Args:
            other (:obj:`Model`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two models are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and self.source == other.source \
            and self.language == other.language \
            and are_lists_equal(self.changes, other.changes)


class ModelChange(abc.ABC):
    """ A change to a model

    Attributes:
        name (:obj:`str`): name
        target (:obj:`str`): path to the model element that should be changed
    """

    def __init__(self, name=None, target=None):
        """
        Args:
            name (:obj:`str`, optional): name
            target (:obj:`str`, optional): path to the model element that should be changed
        """
        self.name = name
        self.target = target

    @abc.abstractmethod
    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        pass  # pragma: no cover

    def is_equal(self, other):
        """ Determine if model changes are equal

        Args:
            other (:obj:`ModelChange`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two model changes are equal
        """
        return self.__class__ == other.__class__ \
            and self.name == other.name \
            and self.target == other.target


class ModelAttributeChange(ModelChange):
    """ A change of an attribute of a model

    Attributes:
        target (:obj:`str`): path to the model element that should be changed
        new_value (:obj:`str`): new value
    """

    def __init__(self, target=None, new_value=None):
        """
        Args:
            target (:obj:`str`, optional): path to the model element that should be changed
            new_value (:obj:`str`, optional): new value
        """
        super(ModelAttributeChange, self).__init__(target=target)
        self.new_value = new_value

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.target, self.new_value)

    def is_equal(self, other):
        """ Determine if model attribute changes are equal

        Args:
            other (:obj:`ModelAttributeChange`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two model attribute changes are equal
        """
        return super(ModelAttributeChange, self).is_equal(other) \
            and self.new_value == other.new_value


class AbstractTask(abc.ABC):
    """ Base class for tasks

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
    """

    def __init__(self, id=None, name=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
        """
        self.id = id
        self.name = name

    @abc.abstractmethod
    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        pass  # pragma: no cover

    def is_equal(self, other):
        """ Determine if tasks are equal

        Args:
            other (:obj:`AbstractTask`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two tasks are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name


class Task(AbstractTask):
    """ A task

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        model (:obj:`Model`): model
        simulation (:obj:`Simulation`): simulation
    """

    def __init__(self, id=None, name=None, model=None, simulation=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            model (:obj:`Model`, optional): model
            simulation (:obj:`Simulation`, optional): simulation
        """
        super(Task, self).__init__(id=id, name=name)
        self.model = model
        self.simulation = simulation

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name,
                self.model.to_tuple() if self.model else None,
                self.simulation.to_tuple() if self.simulation else None)

    def is_equal(self, other):
        """ Determine if tasks are equal

        Args:
            other (:obj:`Task`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two tasks are equal
        """
        return super(Task, self).is_equal(other) \
            and ((self.model is None and self.model == other.model)
                 or (self.model is not None and self.model.is_equal(other.model))) \
            and ((self.simulation is None and self.simulation == other.simulation)
                 or (self.simulation is not None and self.simulation.is_equal(other.simulation)))


class DataGenerator(object):
    """ A data generator

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        variables (:obj:`list` of :obj:`DataGeneratorVariable`): variables
        parameters (:obj:`list` of :obj:`DataGeneratorParameters`): variables
        math (:obj:`str`): mathematical expression
    """

    def __init__(self, id=None, name=None, variables=None, parameters=None, math=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            variables (:obj:`list` of :obj:`DataGeneratorVariable`, optional): variables
            parameters (:obj:`list` of :obj:`DataGeneratorParameters`, optional): variables
            math (:obj:`str`, optional): mathematical expression
        """
        self.id = id
        self.name = name
        self.variables = variables or []
        self.parameters = parameters or []
        self.math = math

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name,
                tuple(none_sorted(variable.to_tuple() for variable in self.variables)),
                tuple(none_sorted(parameter.to_tuple() for parameter in self.parameters)),
                self.math)

    def is_equal(self, other):
        """ Determine if data generators
         are equal

        Args:
            other (:obj:`DataGenerator`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two data generators are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and are_lists_equal(self.variables, other.variables) \
            and are_lists_equal(self.parameters, other.parameters) \
            and self.math == other.math


class DataGeneratorVariable(object):
    """ A model variable involved in a data generator

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        target (:obj:`str`): target
        symbol (:obj:`str`): symbol
        task (:obj:`AbstractTask`): task
        model (:obj:`Model`): model
    """

    def __init__(self, id=None, name=None, target=None, symbol=None, task=None, model=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            target (:obj:`str`, optional): target
            symbol (:obj:`str`, optional): symbol
            task (:obj:`AbstractTask`, optional): task
            model (:obj:`Model`, optional): model
        """
        self.id = id
        self.name = name
        self.target = target
        self.symbol = symbol
        self.task = task
        self.model = model

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name, self.target, self.symbol,
                self.task.to_tuple() if self.task else None,
                self.model.to_tuple() if self.model else None)

    def is_equal(self, other):
        """ Determine if data generator variables are equal

        Args:
            other (:obj:`DataGeneratorVariable`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two data generator variables are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and self.target == other.target \
            and self.symbol == other.symbol \
            and ((self.task is None and self.task == other.task)
                 or (self.task is not None and self.task.is_equal(other.task))) \
            and ((self.model is None and self.model == other.model)
                 or (self.model is not None and self.model.is_equal(other.model)))


class DataGeneratorParameter(object):
    """ A parameter involved in a data generator

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        value (:obj:`float`): value
    """

    def __init__(self, id=None, name=None, value=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            value (:obj:`float`, optional): value
        """
        self.id = id
        self.name = name
        self.value = value

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name, self.value)

    def is_equal(self, other):
        """ Determine if data generator parameters are equal

        Args:
            other (:obj:`DataGeneratorParameter`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two data generator parameters are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and self.value == other.value


class Output(abc.ABC):
    """ An output

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
    """

    def __init__(self, id=None, name=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
        """
        self.id = id
        self.name = name

    @abc.abstractmethod
    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        pass  # pragma: no cover

    def is_equal(self, other):
        """ Determine if outputs are equal

        Args:
            other (:obj:`Output`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two outputs are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name


class Report(Output):
    """ A report

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        data_sets (:obj:`list` of :obj:`DataSet`): data sets
    """

    def __init__(self, id=None, name=None, data_sets=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            data_sets (:obj:`list` of :obj:`DataSet`, optional): data sets
        """
        super(Report, self).__init__(id=id, name=name)
        self.data_sets = data_sets or []

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name,
                tuple(none_sorted(data_set.to_tuple() for data_set in self.data_sets)))

    def is_equal(self, other):
        """ Determine if reports are equal

        Args:
            other (:obj:`Report`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two reports are equal
        """
        return super(Report, self).is_equal(other) \
            and are_lists_equal(self.data_sets, other.data_sets)


class DataSet(object):
    """ A data set in a report

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        label (:obj:`str`): label
        data_generator (:obj:`DataGenerator`): data generator
    """

    def __init__(self, id=None, name=None, label=None, data_generator=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            label (:obj:`str`, optional): label
            data_generator (:obj:`DataGenerator`, optional): data generator
        """
        self.id = id
        self.name = name
        self.label = label
        self.data_generator = data_generator

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name, self.label,
                self.data_generator.to_tuple() if self.data_generator is not None else None)

    def is_equal(self, other):
        """ Determine if data sets are equal

        Args:
            other (:obj:`DataSet`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two data sets are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and self.label == other.label \
            and ((self.data_generator is None and self.data_generator == other.data_generator)
                 or (self.data_generator is not None and self.data_generator.is_equal(other.data_generator))) \



class Plot2D(Output):
    """ A 2D plot

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        curves (:obj:`list` of :obj:`Curve`): curves
    """

    def __init__(self, id=None, name=None, curves=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            curves (:obj:`list` of :obj:`Curve`, optional): curves
        """
        self.id = id
        self.name = name
        self.curves = curves or []

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name,
                tuple(none_sorted(curve.to_tuple() for curve in self.curves)))

    def is_equal(self, other):
        """ Determine if plots are equal

        Args:
            other (:obj:`Plot2D`): another plot

        Returns:
            :obj:`bool`: :obj:`True`, if two plots are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and are_lists_equal(self.curves, other.curves)


class Plot3D(Output):
    """ A 3D plot

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        surfaces (:obj:`list` of :obj:`Surface`): surfaces
    """

    def __init__(self, id=None, name=None, surfaces=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            surfaces (:obj:`list` of :obj:`Surface`, optional): surfaces
        """
        self.id = id
        self.name = name
        self.surfaces = surfaces or []

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name,
                tuple(none_sorted(surface.to_tuple() for surface in self.surfaces)))

    def is_equal(self, other):
        """ Determine if plots are equal

        Args:
            other (:obj:`Plot3D`): another plot

        Returns:
            :obj:`bool`: :obj:`True`, if two plots are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and are_lists_equal(self.surfaces, other.surfaces)


class AxisScale(str, enum.Enum):
    """ Scale of an axis """
    linear = 'linear'
    log = 'log'


class Curve(object):
    """ A curve in a 2D plot

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        x_scale (:obj:`AxisScale`): x axis scale
        y_scale (:obj:`AxisScale`): y axis scale
        x_data_generator (:obj:`DataGenerator`): x data generator
        y_data_generator (:obj:`DataGenerator`): y data generator
    """

    def __init__(self, id=None, name=None, x_scale=None, y_scale=None, x_data_generator=None, y_data_generator=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            x_scale (:obj:`AxisScale`, optional): x axis scale
            y_scale (:obj:`AxisScale`, optional): y axis scale
            x_data_generator (:obj:`DataGenerator`, optional): x data generator
            y_data_generator (:obj:`DataGenerator`, optional): y data generator
        """
        self.id = id
        self.name = name
        self.x_scale = x_scale
        self.y_scale = y_scale
        self.x_data_generator = x_data_generator
        self.y_data_generator = y_data_generator

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name,
                self.x_scale.value if self.x_scale else None,
                self.y_scale.value if self.y_scale else None,
                self.x_data_generator.to_tuple() if self.x_data_generator else None,
                self.y_data_generator.to_tuple() if self.y_data_generator else None)

    def is_equal(self, other):
        """ Determine if curves are equal

        Args:
            other (:obj:`Curve`): another curve

        Returns:
            :obj:`bool`: :obj:`True`, if two curves are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and self.x_scale == other.x_scale \
            and self.y_scale == other.y_scale \
            and ((self.x_data_generator is None and self.x_data_generator == other.x_data_generator)
                 or (self.x_data_generator is not None and self.x_data_generator.is_equal(other.x_data_generator))) \
            and ((self.y_data_generator is None and self.y_data_generator == other.y_data_generator)
                 or (self.y_data_generator is not None and self.y_data_generator.is_equal(other.y_data_generator)))


class Surface(object):
    """ A surface in a 3D plot

    Attributes:
        id (:obj:`str`): id
        name (:obj:`str`): name
        x_scale (:obj:`AxisScale`): x axis scale
        y_scale (:obj:`AxisScale`): y axis scale
        z_scale (:obj:`AxisScale`): z axis scale
        x_data_generator (:obj:`DataGenerator`): x data generator
        y_data_generator (:obj:`DataGenerator`): y data generator
        z_data_generator (:obj:`DataGenerator`): z data generator
    """

    def __init__(self, id=None, name=None,
                 x_scale=None, y_scale=None, z_scale=None,
                 x_data_generator=None, y_data_generator=None, z_data_generator=None):
        """
        Args:
            id (:obj:`str`, optional): id
            name (:obj:`str`, optional): name
            x_scale (:obj:`AxisScale`, optional): x axis scale
            y_scale (:obj:`AxisScale`, optional): y axis scale
            z_scale (:obj:`AxisScale`, optional): z axis scale
            x_data_generator (:obj:`DataGenerator`, optional): x data generator
            y_data_generator (:obj:`DataGenerator`, optional): y data generator
            z_data_generator (:obj:`DataGenerator`, optional): z data generator
        """
        self.id = id
        self.name = name
        self.x_scale = x_scale
        self.y_scale = y_scale
        self.z_scale = z_scale
        self.x_data_generator = x_data_generator
        self.y_data_generator = y_data_generator
        self.z_data_generator = z_data_generator

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.id, self.name,
                self.x_scale.value if self.x_scale else None,
                self.y_scale.value if self.y_scale else None,
                self.z_scale.value if self.z_scale else None,
                self.x_data_generator.to_tuple() if self.x_data_generator else None,
                self.y_data_generator.to_tuple() if self.y_data_generator else None,
                self.z_data_generator.to_tuple() if self.z_data_generator else None)

    def is_equal(self, other):
        """ Determine if surfaces are equal

        Args:
            other (:obj:`Surface`): another surface

        Returns:
            :obj:`bool`: :obj:`True`, if two surfaces are equal
        """
        return self.__class__ == other.__class__ \
            and self.id == other.id \
            and self.name == other.name \
            and self.x_scale == other.x_scale \
            and self.y_scale == other.y_scale \
            and self.z_scale == other.z_scale \
            and ((self.x_data_generator is None and self.x_data_generator == other.x_data_generator)
                 or (self.x_data_generator is not None and self.x_data_generator.is_equal(other.x_data_generator))) \
            and ((self.y_data_generator is None and self.y_data_generator == other.y_data_generator)
                 or (self.y_data_generator is not None and self.y_data_generator.is_equal(other.y_data_generator))) \
            and ((self.z_data_generator is None and self.z_data_generator == other.z_data_generator)
                 or (self.z_data_generator is not None and self.z_data_generator.is_equal(other.z_data_generator)))



""" Data model for reports

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import enum


__all__ = ['DataGeneratorVariableResults', 'OutputResults', 'ReportFormat']


class DataGeneratorVariableResults(dict):
    """ Dictionary that maps the ids of variables (:obj:`DataGeneratorVariable`) to their results (:obj:`numpy.ndarray`)

    * Keys (:obj:`str`): ids of variables (:obj:`DataGeneratorVariable`)
    * Values (:obj:`numpy.ndarray`): result of each variable

        * Steady-state tasks of non-spatial models: results should be arrays of shape ``(1,)``
        * One-step tasks of non-spatial models: results should be arrays of shape ``(2,)``
        * Uniform time course tasks of non-spatial models: results should be arrays of shape ``(number_of_points + 1,)``
    """
    pass


class OutputResults(dict):
    """ Dictionary that maps the ids of outputs (e.g., :obj:`Report`) to their results (:obj:`pandas.DataFrame`)

    * Keys (:obj:`str`): ids of outputs (e.g., :obj:`Report`)
    * Values (:obj:`pandas.DataFrame`): result of each output

        * Data:

            * Steady-state tasks of non-spatial models: results should be arrays of shape ``(number of data sets, 1)``
            * One-step tasks of non-spatial models: results should be arrays of shape ``(number of data sets, 2)``
            * Uniform time course tasks of non-spatial models: results should be arrays of shape ``(number of data sets, number_of_points + 1)``

        * Indices (row labels)

            * Reports: equal to the ids of the data sets if each report
    """
    pass


class ReportFormat(str, enum.Enum):
    """ Format of a report """
    csv = 'csv'
    h5 = 'h5'
    hdf = 'h5'
    hdf5 = 'h5'



""" Utilities for reading the specifications of a simulator

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..sedml.data_model import Algorithm, AlgorithmParameterChange
from .data_model import SoftwareInterface
import json


def gen_algorithms_from_specs(specifications):
    """ Generate a list of algorithms and their parameters from the specifications of a simulator

    Args:
        specifications (:obj:`dict` or :obj:`str`): specifications or path to specifications

    Returns:
        :obj:`dict` of :obj:`str` to :obj:`Algorithm`: dictionary that maps KiSAO ids to algorithms and their parameters
    """
    if isinstance(specifications, str):
        with open(specifications, 'r') as file:
            specifications = json.load(file)

    algs = {}
    for alg_spec in specifications.get('algorithms', []):
        if SoftwareInterface.biosimulators_docker_image.value in alg_spec['availableSoftwareInterfaceTypes']:

            alg = Algorithm()
            alg.kisao_id = alg_spec.get('kisaoId', {}).get('id', None)
            algs[alg.kisao_id] = alg

            param_specs = alg_spec.get('parameters', None)
            if param_specs:
                for param_spec in param_specs:
                    if SoftwareInterface.biosimulators_docker_image.value in param_spec['availableSoftwareInterfaceTypes']:
                        alg.changes.append(AlgorithmParameterChange(
                            kisao_id=param_spec.get('kisaoId', {}).get('id', None),
                            new_value=param_spec.get('value', None),
                        ))

    return algs



""" Data model for BioSimulations metadata

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..data_model import Person, Identifier, OntologyTerm  # noqa: F401
from ..utils.core import are_lists_equal, none_sorted
import datetime  # noqa: F401

__all__ = [
    'Metadata',
    'ExternalReferences',
    'Citation',
]


class Metadata(object):
    """ Metadata about an object

    Attributes:
        description (:obj:`str`): description
        tags (:obj:`list` of :obj:`str`): tags
        authors (:obj:`list` of :obj:`Person`): authors
        references (:obj:`ExternalReferences`): identifiers and citations
        license (:obj:`OntologyTerm`): license
        created (:obj:`datetime.datetime`): created date-time
        updated (:obj:`datetime.datetime`): updated date-time
    """

    def __init__(self, description=None, tags=None, authors=None, references=None, license=None, created=None, updated=None):
        """
        Args:
            description (:obj:`str`, optional): description
            tags (:obj:`list` of :obj:`str`, optional): tags
            authors (:obj:`list` of :obj:`Person`, optional): authors
            references (:obj:`ExternalReferences`, optional): identifiers and citations
            license (:obj:`OntologyTerm`, optional): license
            created (:obj:`datetime.datetime`, optional): created date-time
            updated (:obj:`datetime.datetime`, optional): updated date-time
        """
        self.description = description
        self.tags = tags or []
        self.authors = authors or []
        self.references = references
        self.license = license
        self.created = created
        self.updated = updated

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (self.description, tuple(self.tags),
                tuple(none_sorted(author.to_tuple() for author in self.authors)),
                self.references.to_tuple() if self.references else None,
                self.license.to_tuple() if self.license else None,
                self.created,
                self.updated,
                )

    def is_equal(self, other):
        """ Determine if metadata are equal

        Args:
            other (:obj:`Metadata`): another metadata

        Returns:
            :obj:`bool`: :obj:`True`, if two metadata are equal
        """
        return self.__class__ == other.__class__ \
            and self.description == other.description \
            and none_sorted(self.tags) == none_sorted(other.tags) \
            and are_lists_equal(self.authors, other.authors) \
            and ((self.references is None and self.references == other.references)
                 or (self.references is not None and self.references.is_equal(other.references))) \
            and ((self.license is None and self.license == other.license)
                 or (self.license is not None and self.license.is_equal(other.license))) \
            and self.created == other.created \
            and self.updated == other.updated


class ExternalReferences(object):
    """ Identifiers and citations of an object

    Attributes:
        identifiers (:obj:`list` of :obj:`Identifier`): identifiers
        citations (:obj:`list` of :obj:`Citation`): citations
    """

    def __init__(self, identifiers=None, citations=None):
        """
        Args:
            identifiers (:obj:`list` of :obj:`Identifier`, optional): identifiers
            citations (:obj:`list` of :obj:`Citation`, optional): citations
        """
        self.identifiers = identifiers or []
        self.citations = citations or []

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (
            tuple(none_sorted(identifier.to_tuple() for identifier in self.identifiers)),
            tuple(none_sorted(citation.to_tuple() for citation in self.citations)),
        )

    def is_equal(self, other):
        """ Determine if collections of external references are equal

        Args:
            other (:obj:`ExternalReferences`): another collection of external referencse

        Returns:
            :obj:`bool`: :obj:`True`, if two collections of external references are equal
        """
        return self.__class__ == other.__class__ \
            and are_lists_equal(self.identifiers, other.identifiers) \
            and are_lists_equal(self.citations, other.citations)


class Citation(object):
    """ A citation

    Attributes:
        title (:obj:`str`): title
        authors (:obj:`str`): authors
        journal (:obj:`str`): journal
        volume (:obj:`str`): volume
        issue (:obj:`str`): issue
        pages (:obj:`str`): pages
        year (:obj:`int`): year
        identifiers (:obj:`list` of :obj:`Identifier`): identifiers
    """

    def __init__(self, title=None, authors=None, journal=None, volume=None, issue=None, pages=None, year=None, identifiers=None):
        """
        Args:
            title (:obj:`str`, optional): title
            authors (:obj:`str`, optional): authors
            journal (:obj:`str`, optional): journal
            volume (:obj:`str`, optional): volume
            issue (:obj:`str`, optional): issue
            pages (:obj:`str`, optional): pages
            year (:obj:`int`, optional): year
            identifiers (:obj:`list` of :obj:`Identifier`, optional): identifiers
        """
        self.title = title
        self.authors = authors
        self.journal = journal
        self.volume = volume
        self.issue = issue
        self.pages = pages
        self.year = year
        self.identifiers = identifiers or []

    def to_tuple(self):
        """ Get a tuple representation

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation
        """
        return (
            self.title,
            self.authors,
            self.journal,
            self.volume,
            self.issue,
            self.pages,
            self.year,
            tuple(none_sorted(identifier.to_tuple() for identifier in self.identifiers)),
        )

    def is_equal(self, other):
        """ Determine if citations are equal

        Args:
            other (:obj:`Citation`): another citation

        Returns:
            :obj:`bool`: :obj:`True`, if two citations are equal
        """
        return self.__class__ == other.__class__ \
            and self.title == other.title \
            and self.authors == other.authors \
            and self.journal == other.journal \
            and self.volume == other.volume \
            and self.issue == other.issue \
            and self.pages == other.pages \
            and self.year == other.year \
            and are_lists_equal(self.identifiers, other.identifiers)



""" Data model for COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..utils.core import are_lists_equal, none_sorted
from ..data_model import Person  # noqa: F401
import abc
import datetime  # noqa: F401
import enum

__all__ = [
    'CombineArchiveBase',
    'CombineArchive',
    'CombineArchiveContent',
    'CombineArchiveContentFormat',
    'CombineArchiveContentFormatPattern',
]


class CombineArchiveBase(abc.ABC):
    """ A COMBINE/OMEX archive

    Attributes:
        description (:obj:`str`): description
        authors (:obj:`list` of :obj:`Person`): authors
        created (:obj:`datetime.datetime`): created date
        updated (:obj:`datetime.datetime`): updated date
    """
    pass


class CombineArchive(CombineArchiveBase):
    """ A COMBINE/OMEX archive

    Attributes:
        contents (:obj:`list` of :obj:`CombineArchiveContent`): contents of the archive
        description (:obj:`str`): description
        authors (:obj:`list` of :obj:`Person`): authors
        created (:obj:`datetime.datetime`): created date
        updated (:obj:`datetime.datetime`): updated date
    """

    def __init__(self, contents=None, description=None, authors=None, created=None, updated=None):
        """
        Args:
            contents (:obj:`list` of :obj:`CombineArchiveContent`, optional): contents of the archive
            description (:obj:`str`, optional): description
            authors (:obj:`list` of :obj:`Person`, optional): authors
            created (:obj:`datetime.datetime`, optional): created date
            updated (:obj:`datetime.datetime`, optional): updated date
        """
        self.contents = contents or []
        self.description = description
        self.authors = authors or []
        self.created = created
        self.updated = updated

    def get_master_content(self):
        """ Get the master content of an archive

        Returns:
            :obj:`CombineArchiveContent` or :obj:`None`: master content

        Raises:
            :obj:`ValueError`: if more than one content item is marked as master
        """
        master_content = []
        for content in self.contents:
            if content.master:
                master_content.append(content)
        if not master_content:
            return None
        if len(master_content) == 1:
            return master_content[0]
        raise ValueError('Multiple content items are marked as master')

    def to_tuple(self):
        """ Tuple representation of a COMBINE/OMEX archive

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation of a COMBINE/OMEX archive
        """
        contents = tuple(none_sorted(content.to_tuple() for content in self.contents))
        authors = tuple(none_sorted(author.to_tuple() for author in self.authors))
        return (contents, self.description, authors, self.created, self.updated)

    def is_equal(self, other):
        """ Determine if two content items are equal

        Args:
            other (:obj:`CombineArchiveContent`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two archives are equal
        """
        return self.__class__ == other.__class__ \
            and are_lists_equal(self.contents, other.contents) \
            and self.description == other.description \
            and are_lists_equal(self.authors, other.authors) \
            and self.created == other.created \
            and self.updated == other.updated


class CombineArchiveContent(CombineArchiveBase):
    """ A content item (e.g., file) in a COMBINE/OMEX archive

    Attributes:
        location (:obj:`str`): path to the content
        format (:obj:`str`): URL for the specification of the format of the content
        master (:obj:`bool`): :obj:`True`, if the content is the "primary" content of the parent archive
        description (:obj:`str`): description
        authors (:obj:`list` of :obj:`Person`): authors
        created (:obj:`datetime.datetime`): created date
        updated (:obj:`datetime.datetime`): updated date
    """

    def __init__(self, location=None, format=None, master=False, description=None, authors=None, created=None, updated=None):
        """
        Args:
            location (:obj:`str`, optional): path to the content
            format (:obj:`str`, optional): URL for the specification of the format of the content
            master (:obj:`bool`, optional): :obj:`True`, if the content is the "primary" content of the parent archive
            description (:obj:`str`, optional): description
            authors (:obj:`list` of :obj:`Person`, optional): authors
            created (:obj:`datetime.datetime`, optional): created date
            updated (:obj:`datetime.datetime`, optional): updated date
        """
        self.location = location
        self.format = format
        self.master = master
        self.description = description
        self.authors = authors or []
        self.created = created
        self.updated = updated

    def to_tuple(self):
        """ Tuple representation of a content item of a COMBINE/OMEX archive

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation of a content item of a COMBINE/OMEX archive
        """
        authors = tuple(none_sorted(author.to_tuple() for author in self.authors))
        return (self.location, self.format, self.master, self.description, authors, self.created, self.updated)

    def is_equal(self, other):
        """ Determine if two content items are equal

        Args:
            other (:obj:`CombineArchiveContent`): another content item

        Returns:
            :obj:`bool`: :obj:`True`, if two content items are equal
        """
        return self.__class__ == other.__class__ \
            and self.location == other.location \
            and self.format == other.format \
            and self.master == other.master \
            and self.description == other.description \
            and are_lists_equal(self.authors, other.authors) \
            and self.created == other.created \
            and self.updated == other.updated


class CombineArchiveContentFormat(str, enum.Enum):
    """ Format for the content of COMBINE/OMEX archives """
    BNGL = 'http://bionetgen.org/'
    CellML = 'http://identifiers.org/combine.specifications/cellml'
    SBGN = 'http://identifiers.org/combine.specifications/sbgn'
    SBML = 'http://identifiers.org/combine.specifications/sbml'
    SED_ML = 'http://identifiers.org/combine.specifications/sed-ml'


class CombineArchiveContentFormatPattern(str, enum.Enum):
    """ Format for the content of COMBINE/OMEX archives """
    BNGL = r'^https?://bionetgen\.org($|/)'
    CellML = r'^https?://identifiers\.org/combine\.specifications/cellml($|\.)'
    SBGN = r'^https?://identifiers\.org/combine\.specifications/sbgn($|\.)'
    SBML = r'^https?://identifiers\.org/combine\.specifications/sbml(\.level\-\d+\.version\-\d+)?$'
    SED_ML = r'^https?://identifiers\.org/combine\.specifications/sed\-?ml(\.level\-\d+\.version\-\d+)?$'



""" Utilities for logging

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import logging
import os

__all__ = ['get_logger']

FILENAME_PATTERN = '~/.biosimulators-utils/logs/{}.log'


def get_logger(name='log', filename_pattern=FILENAME_PATTERN):
    """ Get a logger

    Args:
        name (:obj:`str`, optional): name
        filename_pattern (:obj:`str`, optional): pattern for paths for log files

    Returns:
        :obj:`logging.Logger`: logger
    """
    logger = logging.getLogger('biosimulators-utils-' + name)
    if not logger.handlers:
        config_logger(logger, name, filename_pattern=filename_pattern)
    return logger


def config_logger(logger, name='log', filename_pattern=FILENAME_PATTERN):
    """ Configure a new logger

    Args:
        logger (:obj:`logging.Logger`): logger
        name (:obj:`str`, optional): name
        filename_pattern (:obj:`str`, optional): pattern for paths for log files
    """
    formatter = logging.Formatter(
        '%(levelname)s'
        '-%(asctime)s'
        '-%(pathname)s'
        '-%(funcName)s'
        '-%(lineno)d'
        ': %(message)s'
    )

    log_filename = os.path.expanduser(filename_pattern.format(name))
    log_dir = os.path.dirname(log_filename)
    if not os.path.isdir(log_dir):
        os.makedirs(log_dir)
    handler = logging.FileHandler(log_filename)
    handler.setFormatter(formatter)

    logger = logging.getLogger('biosimulators-utils-' + name)
    logger.addHandler(handler)

    logger.log(logging.INFO, (
        '\n'
        '\n'
        '===============\n'
        'Log initialized\n'
        '==============='
    ))



""" Data model for archives (e.g., zip files)

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..utils.core import are_lists_equal, none_sorted

__all__ = [
    'Archive',
    'ArchiveFile',
]


class Archive(object):
    """ An archive (e.g., zip file)

    Attributes:
        files (:obj:`list` of :obj:`ArchiveFile`): files
    """

    def __init__(self, files=None):
        """
        Args:
            files (:obj:`list` of :obj:`ArchiveFile`, optional): files
        """
        self.files = files or []

    def to_tuple(self):
        """ Tuple representation of an archive

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation of an archive
        """
        return (
            tuple(none_sorted(file.to_tuple() for file in self.files)),
        )

    def is_equal(self, other):
        """ Determine if two archives are equal

        Args:
            other (:obj:`ArchiveFile`): another archive

        Returns:
            :obj:`bool`: :obj:`True`, if two archives are equal
        """
        return self.__class__ == other.__class__ \
            and are_lists_equal(self.files, other.files)


class ArchiveFile(object):
    """ A file in a archive (e.g., zip file)

    Attributes:
        local_path (:obj:`str`): path within local filesytem
        archive_path (:obj:`str`): archive_path
    """

    def __init__(self, local_path=None, archive_path=None):
        """
        Args:
            local_path (:obj:`str`, optional): path within local filesytem
            archive_path (:obj:`str`, optional): archive_path
        """
        self.local_path = local_path
        self.archive_path = archive_path

    def to_tuple(self):
        """ Tuple representation of a file

        Returns:
            :obj:`tuple` of :obj:`str`: tuple representation of a file
        """
        return (self.local_path, self.archive_path)

    def is_equal(self, other):
        """ Determine if two files are equal

        Args:
            other (:obj:`ArchiveFile`): another file

        Returns:
            :obj:`bool`: :obj:`True`, if two files are equal
        """
        return self.__class__ == other.__class__ \
            and self.local_path == other.local_path \
            and self.archive_path == other.archive_path



""" Miscellaneous utilities

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..data_model import ValueType, OntologyTerm
import json
import re

__all__ = [
    'are_lists_equal', 'none_sorted', 'assert_exception',
    'validate_value', 'validate_str_value', 'format_value', 'parse_value',
]


def are_lists_equal(a, b):
    """ Determine if two lists are equal, optionally up to the order of the elements

    Args:
        a (:obj:`list`): first list
        b (:obj:`list`): second list

    Returns:
        :obj:`bool`: :obj:`True`, if lists are equal
    """
    if len(a) != len(b):
        return False

    a = none_sorted(a, key=lambda x: x.to_tuple())
    b = none_sorted(b, key=lambda x: x.to_tuple())

    for a_el, b_el in zip(a, b):
        if not a_el.is_equal(b_el):
            return False

    return True


def none_sort_key_gen(key=None):
    class NoneComparator(object):
        def __init__(self, obj):
            if key:
                self.obj = key(obj)
            else:
                self.obj = obj

        def __lt__(self, other):
            return none_comparator(self.obj, other.obj) < 0

        def __gt__(self, other):
            return none_comparator(self.obj, other.obj) > 0

        def __eq__(self, other):
            return none_comparator(self.obj, other.obj) == 0

        def __le__(self, other):
            return none_comparator(self.obj, other.obj) <= 0

        def __ge__(self, other):
            return none_comparator(self.obj, other.obj) >= 0

        def __ne__(self, other):
            return none_comparator(self.obj, other.obj) != 0
    return NoneComparator


def none_comparator(x, y):
    if x == y:
        return 0

    if isinstance(x, tuple) and not isinstance(y, tuple):
        return 1
    if not isinstance(x, tuple) and isinstance(y, tuple):
        return -1
    if isinstance(x, tuple) and isinstance(y, tuple):
        if len(x) < len(y):
            return -1
        if len(x) > len(y):
            return 1

        for x1, y1 in zip(x, y):
            cmp = none_comparator(x1, y1)
            if cmp != 0:
                return cmp

        return 0  # pragma: no cover

    if x is None:
        return -1
    if y is None:
        return 1

    if x < y:
        return -1
    if x > y:
        return 1


def none_sorted(arr, key=None):
    """ Sort an error that contains :obj:`None`

    Args:
        arr (:obj:`list`): array

    Returns:
        :obj:`list`: sorted array
    """
    return sorted(arr, key=none_sort_key_gen(key))


def assert_exception(success, exception):
    """ Raise an error if :obj:`success` is :obj:`False`

    Args:
        success (:obj:`bool`)
        exception (:obj:`Exception`)

    Raises:
        :obj:`Exception`
    """
    if not success:
        raise exception


def validate_value(val, type):
    """ Determine if a value is a valid instance of type :obj:`type`

    Args:
        val (:obj:`object`): value
        type (:obj:`ValueType`): type

    Returns:
        :obj:`bool`: :obj:`True`, if a value is a valid instance of type :obj:`type`
    """
    if type == ValueType.boolean:
        return isinstance(val, bool)
    if type == ValueType.integer:
        return isinstance(val, int) and not isinstance(val, bool)
    if type == ValueType.float:
        return isinstance(val, float)
    if type == ValueType.string:
        return isinstance(val, str)
    if type == ValueType.kisao_id:
        return isinstance(val, OntologyTerm) and val.namespace == 'KISAO' and re.match(r'^KISAO_\d{7}$', val.id) is not None
    if type == ValueType.list:
        return isinstance(val, list)
    if type == ValueType.object:
        return isinstance(val, dict)
    if type == ValueType.any:
        return True
    raise NotImplementedError('Type {} is not supported'.format(type))


def validate_str_value(str_val, type):
    """ Determine if a value is a valid string representation of type :obj:`type`

    Args:
        str_val (:obj:`str`): string representation of a value
        type (:obj:`ValueType`): type

    Returns:
        :obj:`bool`: :obj:`True`, if a value is a valid string representation of type :obj:`type`
    """
    if type == ValueType.boolean:
        return str_val.lower() in ['true', 'false', '0', '1']
    if type == ValueType.integer:
        try:
            int(str_val)
            return True
        except ValueError:
            return False
    if type == ValueType.float:
        try:
            float(str_val)
            return True
        except ValueError:
            return False
    if type == ValueType.string:
        return True
    if type == ValueType.kisao_id:
        return re.match(r'^KISAO_\d{7}$', str_val)
    if type == ValueType.list:
        try:
            return isinstance(json.loads(str_val), list)
        except json.JSONDecodeError:
            return False
    if type == ValueType.object:
        try:
            return isinstance(json.loads(str_val), dict)
        except json.JSONDecodeError:
            return False
    if type == ValueType.any:
        try:
            json.loads(str_val)
            return True
        except json.JSONDecodeError:
            return False
    raise NotImplementedError('Type {} is not supported'.format(type))


def format_value(val, type):
    """ Format a value as a string (e.g., for use with an attribute of an XML object)

    Args:
        val (:obj:`object`): value
        type (:obj:`ValueType`): type

    Returns:
        :obj:`str`: string representation of the value
    """
    if type == ValueType.boolean:
        return str(val).lower()
    if type == ValueType.integer or type == ValueType.float:
        return str(val)
    if type == ValueType.string:
        return val
    if type == ValueType.kisao_id:
        return val.id
    if type == ValueType.list or type == ValueType.object or type == ValueType.any:
        return json.dumps(val)
    raise NotImplementedError('Type {} is not supported'.format(type))


def parse_value(str_val, type):
    """ Parse the string representation of a value (e.g., for use with an attribute of an XML object)

    Args:
        str_val (:obj:`str`): string representation of the value
        type (:obj:`ValueType`): type

    Returns:
        :obj:`object`: Python representation of the value
    """
    if type == ValueType.boolean:
        return str_val.lower() == 'true' or str_val.lower() == '1'
    if type == ValueType.integer:
        return int(str_val)
    if type == ValueType.float:
        return float(str_val)
    if type == ValueType.string:
        return str_val
    if type == ValueType.kisao_id:
        return OntologyTerm(
            namespace='KISAO',
            id=str_val,
            url='https://www.ebi.ac.uk/ols/ontologies/kisao/terms?iri=http%3A%2F%2Fwww.biomodels.net%2Fkisao%2FKISAO%23' + str_val,
        )
    if type == ValueType.list or type == ValueType.object or type == ValueType.any:
        return json.loads(str_val)
    raise NotImplementedError('Type {} is not supported'.format(type))



""" Utilities for working with KiSAO terms

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import re
import warnings

__all__ = ['normalize_kisao_id']


def normalize_kisao_id(id):
    """ Normalize an id for a KiSAO term to the official pattern ``KISAO_\\d{7}``.

    The official id pattern for KiSAO terms is ``KISAO_\\d{7}``. This is often confused with ``KISAO:\\d{7}`` and ``\\d{7}``.
    This function automatically converts these other patterns to the offfical pattern.

    Args:
        id (:obj:`str`): offical KiSAO id with pattern ``KISAO_\\d{7}`` or a variant such as ``KISAO:\\d{7}`` or ``\\d{7}``

    Returns:
        :obj:`str`: normalized KiSAO id that follows the official pattern ``KISAO_\\d{7}``
    """
    unnormalized_id = id

    id = str(id)

    if id.startswith('KISAO:'):
        id = 'KISAO_' + id[6:]

    if re.match(r'\d+', id):
        id = 'KISAO_' + '0' * (7 - len(id)) + id

    if not re.match(r'KISAO_\d{7}', id):
        warnings.warn("'{}' is likely not an id for a KiSAO term".format(unnormalized_id), UserWarning)

    return id



""" Methods for validating SED objects

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-09
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..xml.utils import validate_xpaths_ref_to_unique_objects
from .data_model import (Task, ModelLanguage, ModelChange, ModelAttributeChange,  # noqa: F401
                         Simulation, UniformTimeCourseSimulation, DataGeneratorVariable,
                         Report, Plot2D, Plot3D)
from .utils import append_all_nested_children_to_doc
import copy
import math
import os
import re

__all__ = [
    'validate_doc',
    'validate_reference',
    'validate_task',
    'validate_model_language',
    'validate_model_change_types',
    'validate_simulation_type',
    'validate_uniform_time_course_simulation',
    'validate_data_generator_variables',
]


def validate_doc(doc, validate_semantics=True):
    """ Validate a SED document

    Args:
        doc (:obj:`SedDocument`): SED document
        validate_semantics (:obj:`bool`, optional): if :obj:`True`, check that SED-ML is semantically valid

    Raises:
        :obj:`ValueError`: if document is invalid (e.g., required ids missing or ids not unique)
    """
    # validate that required ids exist and are unique
    for child_type in ('models', 'simulations', 'data_generators', 'tasks', 'outputs'):
        children = getattr(doc, child_type)

        missing_ids = next((True for child in children if not getattr(child, 'id', None)), False)
        if missing_ids:
            raise ValueError('{} must have ids'.format(child_type))

        if validate_semantics:
            repeated_ids = len(set(getattr(child, 'id', None) for child in children)) < len(children)
            if repeated_ids:
                raise ValueError('{} must have unique ids'.format(child_type))

    if validate_semantics:
        # validate the models, simulations, tasks, data generators are children of the SED document
        doc_copy = copy.deepcopy(doc)
        append_all_nested_children_to_doc(doc_copy)

        for child_type in ('models', 'simulations', 'tasks', 'data_generators'):
            if len(getattr(doc, child_type)) != len(getattr(doc_copy, child_type)):
                raise ValueError('{} must be direct children of SED document'.format(child_type))

        # validate that model attribute changes have targets
        for model in doc.models:
            for change in model.changes:
                if isinstance(change, ModelAttributeChange):
                    if not change.target:
                        raise ValueError('Model change attributes must define a target')

        for sim in doc.simulations:
            if sim.algorithm:
                if not sim.algorithm.kisao_id or not re.match(r'^KISAO_\d{7}$', sim.algorithm.kisao_id):
                    raise ValueError('Algorithm of simulation {} has an invalid KiSAO id: {}'.format(sim.id, sim.algorithm.kisao_id))
                for change in sim.algorithm.changes:
                    if not change.kisao_id or not re.match(r'^KISAO_\d{7}$', change.kisao_id):
                        raise ValueError('Algorithm of simulation {} has an invalid KiSAO id: {}'.format(sim.id, sim.algorithm.kisao_id))

        for task in doc.tasks:
            if isinstance(task, Task):
                validate_reference(task, 'Task {}'.format(task.id), 'model', 'model')
                validate_reference(task, 'Task {}'.format(task.id), 'simulation', 'simulation')

        for data_gen in doc.data_generators:
            for var in data_gen.variables:
                if not var.id:
                    raise ValueError('Variables must have ids')
                if (not var.target and not var.symbol) or (var.target and var.symbol):
                    raise ValueError('Variables must define a target or symbol')
                validate_reference(var, 'Variable {} of data generator "{}"'.format(var.id, data_gen.id), 'task', 'task')
                # model reference not validated because its optional in this context

                if var.model and var.task and var.task.model and var.task.model != var.model:
                    raise ValueError('Model of variable {} of data generator "{}" and model of task must be consistent'.format(
                        var.id, data_gen.id))
            if not data_gen.math:
                raise ValueError('Data generators must have math')

        for output in doc.outputs:
            if isinstance(output, Report):
                for data_set in output.data_sets:
                    if not data_set.id:
                        raise ValueError('Data sets must have ids')
                    if not data_set.label:
                        raise ValueError('Data sets must have labels')
                    validate_reference(data_set, 'Data set {} of report "{}"'.format(
                        data_set.id, output.id), 'data_generator', 'data data generator')

            elif isinstance(output, Plot2D):
                for curve in output.curves:
                    if not curve.id:
                        raise ValueError('Curves must have ids')
                    validate_reference(curve, 'Curve {} of 2D plot "{}"'.format(
                        curve.id, output.id), 'x_data_generator', 'x data data generator')
                    validate_reference(curve, 'Curve {} of 2D plot "{}"'.format(
                        curve.id, output.id), 'y_data_generator', 'y data data generator')

            elif isinstance(output, Plot3D):
                for surface in output.surfaces:
                    if not surface.id:
                        raise ValueError('Surfaces must have ids')
                    validate_reference(surface, 'Surface {} of 3D plot "{}"'.format(
                        surface.id, output.id), 'x_data_generator', 'x data data generator')
                    validate_reference(surface, 'Surface {} of 3D plot "{}"'.format(
                        surface.id, output.id), 'y_data_generator', 'y data data generator')
                    validate_reference(surface, 'Surface {} of 3D plot "{}"'.format(
                        surface.id, output.id), 'z_data_generator', 'z data data generator')


def validate_reference(obj, obj_label, attr_name, attr_label):
    if not getattr(obj, attr_name):
        raise ValueError('{} must have a {}'.format(obj_label, attr_label))


def validate_task(task):
    """ Validate a simulation task

    Args:
        task (:obj:`Task`): task

    Raises:
        :obj:`ValueError`: if task is invalid
        :obj:`FileNotFoundError`: if model file doesn't exist
    """
    # check that task is an instance of Task
    if not isinstance(task, Task):
        raise ValueError('Task type {} is not supported'.format(task.__class__.__name__))

    # check that task has model
    if not task.model:
        raise ValueError('Task must have a model')

    # check that model file exists
    if not task.model.source or not os.path.isfile(task.model.source):
        raise FileNotFoundError("Model source '{}' must be a file".format(task.model.source or ''))

    # check that task has model
    simulation = task.simulation
    if not simulation:
        raise ValueError('Task must have a simulation')

    if not simulation.algorithm:
        raise ValueError('Simulation must have an algorithm')

    if not simulation.algorithm.kisao_id or not re.match(r'^KISAO_\d{7}$', simulation.algorithm.kisao_id):
        raise ValueError('Algorithm must have a valid KiSAO id')

    for change in simulation.algorithm.changes:
        if not change.kisao_id or not re.match(r'^KISAO_\d{7}$', change.kisao_id):
            raise ValueError('Algorithm change must have a valid KiSAO id')


def validate_model_language(language, valid_language):
    """ Check that model is encoded in a specific language

    Args:
        language (:obj:`ModelLanguage`): model language
        valid_language (:obj:`ModelLanguage`): valid model language

    Raises:
        :obj:`NotImplementedError`: if the model uses a different language
    """
    if not language or not re.match('^{}($|:)'.format(valid_language.value), language):
        raise NotImplementedError("Model language {} is not supported. Model language must be '{}'.".format(
            language, valid_language.value))


def validate_model_change_types(changes, types):
    """ Check that model changes are valid

    Args:
        changes (:obj:`list` of :obj:`ModelChange`): model changes
        types (:obj:`tuple` of :obj:`type`): valid model change types

    Raises:
        :obj:`NotImplementedError`: if the model uses different types of changes
    """
    for change in changes:
        if not isinstance(change, types):
            raise NotImplementedError("".join([
                'Model changes of type {} are not supported. '.format(change.__class__.__name__),
                'Model changes must be instances of one of of the following types:\n  - {}'.format(
                    '\n  - '.join(type.__name__ for type in types)),

            ]))

        if isinstance(change, ModelAttributeChange):
            if not change.target:
                raise ValueError('Model change attributes must define a target')


def validate_simulation_type(simulation, types):
    """ Check that simulation is a time course simulation

    Args:
        simulation (:obj:`Simulation`): simulation
        types (:obj:`type`): valid simulation types

    Raises:
        :obj:`NotImplementedError`: if the simulation is a different type
    """
    if not isinstance(simulation, types):
        raise NotImplementedError(
            'Simulation type {} is not supported. Simulation must be an instance of one of the following:\n  - {}'.format(
                simulation.__class__.__name__, '\n  - '.join(type.__name__ for type in types)))


def validate_uniform_time_course_simulation(simulation):
    """ Check that simulation is a valid uniform time course simulation

    Args:
        simulation (:obj:`Simulation`): simulation

    Raises:
        :obj:`ValueErorr`: if the simulation is invalid
    """
    if isinstance(simulation, UniformTimeCourseSimulation):
        if simulation.output_start_time < simulation.initial_time:
            raise ValueError('Output start time {} must be at least the initial time {}.'.format(
                simulation.output_start_time, simulation.initial_time))

        if simulation.output_end_time < simulation.output_start_time:
            raise ValueError('Output end time {} must be at least the output start time {}.'.format(
                simulation.output_end_time, simulation.output_start_time))

        if math.floor(simulation.number_of_points) != simulation.number_of_points:
            raise ValueError('Number of points must be an integer.')


def validate_data_generator_variables(variables):
    """ Check variables have a symbol or target

    Args:
        variables (:obj:`list` of :obj:`DataGeneratorVariable`): variables

    Raises:
        :obj:`ValidateError`: if a variable is invalid
    """
    for variable in variables:
        if (variable.symbol and variable.target) or (not variable.symbol and not variable.target):
            raise ValueError('Variable must define a symbol or target')


def validate_data_generator_variable_xpaths(variables, model_source, attr='id'):
    """ Validate that the target of each data generator variable matches one object in
    an XML-encoded model

    Args:
        variables (:obj:`list` of :obj:`DataGeneratorVariable`): variables
        model_source (:obj:`str`): path to XML model file
        attr (:obj:`str`, optional): attribute to get values of

    Returns:
        :obj:`dict` of :obj:`str` to :obj:`str`: dictionary that maps each XPath to the
            value of the attribute of the object in the XML file that matches the XPath
    """
    x_paths = []
    for variable in variables:
        if variable.target:
            x_path = variable.target
            if '/@' in x_path:
                x_path, _, _ = x_path.rpartition('/@')
            x_paths.append(x_path)
    return validate_xpaths_ref_to_unique_objects(model_source, x_paths, attr=attr)



""" Methods for reading specifications of simulators

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config
import json
import os
import requests
import simplejson.errors


def read_simulator_specs(path_or_url):
    """ Read the specifications of a simulator

    Args:
        path_or_url (:obj:`str`): file path or URL for the specifications of a simulator

    Returns:
        :obj:`dict`: specifications of a simulator

    Raises:
        :obj:`requests.RequestException`: if the specifications could not be retrieved
        :obj:`simplejson.errors.JSONDecodeError`: if the specifications are not propertly encoded into JSON
        :obj:`ValueError`: if the specifications are not consistent with the BioSimulators schema
    """
    if os.path.isfile(path_or_url):
        with open(path_or_url, 'r') as file:
            try:
                specs = json.load(file)
            except json.JSONDecodeError as error:
                raise ValueError(''.join([
                    'Simulator specifications from {} could not be parsed. '.format(path_or_url),
                    'Specifications must be encoded into JSON.\n\n  {}'.format(str(error).replace('\n', '\n  ')),
                ]))

    else:
        # download specifications
        response = requests.get(path_or_url)
        try:
            response.raise_for_status()
        except requests.RequestException as error:
            raise requests.RequestException('Simulator specifications could not be retrieved from {}.\n\n  {}'.format(
                path_or_url, str(error).replace('\n', '\n  ')))

        # check that specifications is valid JSON
        try:
            specs = response.json()
        except simplejson.errors.JSONDecodeError as error:
            raise ValueError(''.join([
                'Simulator specifications from {} could not be parsed. '.format(path_or_url),
                'Specifications must be encoded into JSON.\n\n  {}'.format(str(error).replace('\n', '\n  ')),
            ]))

    # validate specifications
    api_endpoint = get_config().BIOSIMULATORS_API_ENDPOINT
    response = requests.post('{}simulators/validate'.format(api_endpoint), json=specs)
    try:
        response.raise_for_status()
    except requests.RequestException as error:
        raise ValueError(''.join([
            'Simulator specifications from {} are not valid. '.format(path_or_url),
            'Specifications must be adhere to the BioSimulators schema. Documentation is available at {}.\n\n  {}'.format(
                api_endpoint, str(error).replace('\n', '\n  ')),
        ]))

    # return validated specifications
    return specs



""" Utilities for creating zip archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import Archive, ArchiveFile
import os
import zipfile


__all__ = ['ArchiveWriter', 'ArchiveReader']


class ArchiveWriter(object):
    """ Class for writing zip archives """

    def run(self, archive, archive_filename):
        """ Bundle a list of files into a zip archive

        Args:
            archive (:obj:`Archive`): files to bundle into a zip archive
            archive_filename (:obj:`str`): path to save zip file
        """
        with zipfile.ZipFile(archive_filename, mode='w', compression=zipfile.ZIP_LZMA) as zip_file:
            for file in archive.files:
                zip_file.write(file.local_path, arcname=file.archive_path)


class ArchiveReader(object):
    """ Class for reading zip archives """

    def run(self, archive_filename, out_dir=None):
        """ Unpack the files in a zip archive

        Args:
            archive_filename (:obj:`str`): path to zip file
            out_dir (:obj:`str`, optional): path to unpack files

        Returns:
            archive (:obj:`Archive`): files unbundled from zip archive
        """
        archive = Archive()
        with zipfile.ZipFile(archive_filename, mode='r') as zip_file:
            for member in zip_file.namelist():
                archive.files.append(ArchiveFile(
                    archive_path=member,
                    local_path=os.path.join(out_dir, member) if out_dir else None,
                ))
            if out_dir:
                zip_file.extractall(out_dir)
        return archive



""" Utilities for reading and writing COMBINE/OMEX archives

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import CombineArchiveBase, CombineArchive, CombineArchiveContent  # noqa: F401
from ..data_model import Person
import dateutil.parser
import libcombine
import os


__all__ = [
    'CombineArchiveWriter',
    'CombineArchiveReader',
]


class CombineArchiveWriter(object):
    """ Writer for COMBINE/OMEX archives """

    @classmethod
    def run(cls, archive, in_dir, out_file):
        """ Write an archive to a file

        Args:
            archive (:obj:`CombineArchive`): description of archive
            in_dir (:obj:`str`): directory which contains the files in the archive
            out_file (:obj:`str`): path to save archive

        Raises:
            :obj:`AssertionError`: if files could not be added to the archive or the archive could not be
                saved
        """
        # instantiate archive
        archive_comb = libcombine.CombineArchive()

        # set metadata about archive
        cls._write_metadata(archive, archive_comb, '.')

        # add files to archive
        for content in archive.contents:
            assert archive_comb.addFile(
                os.path.join(in_dir, content.location),
                content.location,
                content.format if content.format else '',
                content.master
            )
            cls._write_metadata(content, archive_comb, content.location)

        # save archive to a file
        assert archive_comb.writeToFile(out_file)

    @classmethod
    def _write_metadata(cls, obj, archive_comb, filename):
        """ Write metadata about an archive or a file in an archive

        Args:
            obj (:obj:`CombineArchiveBase`): archive or file in an archive
            archive_comb (:obj:`libcombine.CombineArchive`): archive
            filename (:obj:`str`): path of object with archive

        Raises:
            :obj:`NotImplementedError`: of the updated date is not defined
        """
        desc_comb = libcombine.OmexDescription()
        desc_comb.setAbout(filename)
        if obj.description:
            desc_comb.setDescription(obj.description)
        for author in obj.authors:
            creator_comb = libcombine.VCard()
            if author.given_name:
                creator_comb.setGivenName(author.given_name)
            if author.family_name:
                creator_comb.setFamilyName(author.family_name)
            desc_comb.addCreator(creator_comb)
        if obj.created:
            date_comb = libcombine.Date()
            date_comb.setDateAsString(obj.created.strftime('%Y-%m-%dT%H:%M:%SZ'))
            desc_comb.setCreated(date_comb)
        if obj.updated:
            date_comb = libcombine.Date()
            date_comb.setDateAsString(obj.updated.strftime('%Y-%m-%dT%H:%M:%SZ'))
            desc_comb.getModified().append(date_comb)
        else:
            raise NotImplementedError('libcombine does not support undefined updated dates')
        archive_comb.addMetadata(filename, desc_comb)


class CombineArchiveReader(object):
    """ Reader for COMBINE/OMEX archives """

    NONE_DATETIME = '2000-01-01T00:00:00Z'

    @classmethod
    def run(cls, in_file, out_dir):
        """ Read an archive from a file

        Args:
            in_file (:obj:`str`): path to save archive
            out_dir (:obj:`str`): directory which contains the files in the archive

        Returns:
            :obj:`CombineArchive`: description of archive

        Raises:
            :obj:`ArchiveIoError`: archive is invalid
        """
        archive_comb = libcombine.CombineArchive()
        if not archive_comb.initializeFromArchive(in_file):
            raise ValueError("Invalid COMBINE archive")

        # instantiate archive
        archive = CombineArchive()

        # read metadata
        cls._read_metadata(archive_comb, '.', archive)

        # read files
        for location in archive_comb.getAllLocations():
            location = location.c_str()
            file_comb = archive_comb.getEntryByLocation(location)

            if file_comb.isSetFormat():
                format = file_comb.getFormat()
            else:
                format = None

            content = CombineArchiveContent(
                location=location,
                format=format,
            )
            cls._read_metadata(archive_comb, location, content)
            archive.contents.append(content)

        file_comb = archive_comb.getMasterFile()
        if file_comb:
            location = file_comb.getLocation()
            master_content = next((content for content in archive.contents if content.location == location), None)
            if master_content:
                master_content.master = True

        # extract files
        archive_comb.extractTo(out_dir)

        # return information about archive
        return archive

    @classmethod
    def _read_metadata(cls, archive_comb, filename, obj):
        """ Read metadata about an archive or a file in an archive

        Args:
            archive_comb (:obj:`libcombine.CombineArchive`): archive
            filename (:obj:`str`): path to object within archive
            obj (:obj:`CombineArchiveBase`): object to add metadata to
        """
        desc_comb = archive_comb.getMetadataForLocation(filename)
        if not desc_comb.isEmpty():
            obj.description = desc_comb.getDescription() or None

            for creator_comb in desc_comb.getCreators():
                obj.authors.append(Person(
                    given_name=creator_comb.getGivenName() or None,
                    family_name=creator_comb.getFamilyName() or None,
                ))

            created_comb = desc_comb.getCreated().getDateAsString()
            if created_comb == cls.NONE_DATETIME:
                obj.created = None
            else:
                obj.created = dateutil.parser.parse(created_comb)

            obj.updated = None
            for modified_comb in desc_comb.getModified():
                updated = dateutil.parser.parse(modified_comb.getDateAsString())
                if obj.updated:
                    obj.updated = max(obj.updated, updated)
                else:
                    obj.updated = updated



""" Utilities for reading and writing reports

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config
from .data_model import ReportFormat
import glob
import os
import pandas
import tables

__all__ = [
    'ReportWriter',
    'ReportReader',
]


class ReportWriter(object):
    """ Class for writing reports of simulation results """

    def run(self, results, base_path, rel_path, format=ReportFormat.h5):
        """ Save a report

        Args:
            results (:obj:`pandas.DataFrame`): report results
            base_path (:obj:`str`): path to save results

                * CSV: parent directory to save results
                * HDF5: file to save results

            rel_path (:obj:`str`): path to save results relative to :obj:`base_path`

                * CSV: relative path to :obj:`base_path`
                * HDF5: key within HDF5 file

            format (:obj:`ReportFormat`, optional): report format
        """

        if format == ReportFormat.csv:
            filename = os.path.join(base_path, rel_path + '.' + format.value)
            out_dir = os.path.dirname(filename)
            if not os.path.isdir(out_dir):
                os.makedirs(out_dir)
            results.to_csv(filename,
                           header=False)

        elif format == ReportFormat.h5:
            filename = os.path.join(base_path, get_config().H5_REPORTS_PATH)
            if not os.path.isdir(base_path):
                os.makedirs(base_path)
            results.to_hdf(filename,
                           key=rel_path,
                           format='table',
                           complevel=9,
                           complib='zlib',
                           mode='a',
                           append=False,
                           )

        else:
            raise NotImplementedError('Report format {} is not supported'.format(format))


class ReportReader(object):
    """ Class for reading reports of simulation results """

    def run(self, base_path, rel_path, format=ReportFormat.h5):
        """ Read a report for a file

        Args:
            base_path (:obj:`str`): path to save results

                * CSV: parent directory to save results
                * HDF5: file to save results

            rel_path (:obj:`str`): path to save results relative to :obj:`base_path`

                * CSV: relative path to :obj:`base_path`
                * HDF5: key within HDF5 file

            format (:obj:`ReportFormat`, optional): report format

        Returns:
            :obj:`pandas.DataFrame`: report results
        """
        if format == ReportFormat.csv:
            filename = os.path.join(base_path, rel_path + '.' + format.value)
            df = pandas.read_csv(filename,
                                 index_col=0,
                                 header=None)
            df.columns = pandas.RangeIndex(start=0, stop=df.shape[1], step=1)
            return df

        elif format == ReportFormat.h5:
            filename = os.path.join(base_path, get_config().H5_REPORTS_PATH)
            return pandas.read_hdf(filename,
                                   key=rel_path,
                                   )

        else:
            raise NotImplementedError('Report format {} is not supported'.format(format))

    def get_ids(self, base_path, format=ReportFormat.h5):
        """ Get the ids of the reports in a file

        Args:
            base_path (:obj:`str`): path to save results

                * CSV: parent directory to save results
                * HDF5: file to save results

            format (:obj:`ReportFormat`, optional): report format

        Returns:
            :obj:`set` of :obj:`str`: ids of reports
        """
        if format == ReportFormat.csv:
            report_ids = set()
            for path in glob.glob(os.path.join(base_path, '**/*.' + format.value), recursive=True):
                report_ids.add(os.path.relpath(path, base_path)[0:-len(format.value)-1])
            return report_ids

        elif format == ReportFormat.h5:
            filename = os.path.join(base_path, get_config().H5_REPORTS_PATH)
            reports_file = tables.open_file(filename, mode="r")
            report_ids = set()
            for node in reports_file.walk_nodes():
                base_path, _, rel_path = node._v_pathname.rpartition('/')
                if rel_path == 'table':
                    report_ids.add(base_path[1:])
            reports_file.close()
            return report_ids

        else:
            raise NotImplementedError('Report format {} is not supported'.format(format))



import os


class Config(object):
    """ Configuration

    Attributes:
        REPORT_FORMATS (:obj:`list` of :obj:`str`): default formats to generate reports in
        PLOT_FORMATS (:obj:`list` of :obj:`str`): default formats to generate plots in
        H5_REPORTS_PATH (:obj:`str`): path to save reports in HDF5 format relative to base output directory
        REPORTS_PATH (:obj:`str`): path to save zip archive of reports relative to base output directory
        PLOTS_PATH (:obj:`str`): path to save zip archive of plots relative to base output directory
        BUNDLE_OUTPUTS (:obj:`bool`): indicates whether bundles of report and plot outputs should be produced
        KEEP_INDIVIDUAL_OUTPUTS (:obj:`bool`): indicates whether the individual output files should be kept
        BIOSIMULATORS_API_ENDPOINT (:obj:`str`): URL for BioSimulators API
    """

    def __init__(self, REPORT_FORMATS, PLOT_FORMATS,
                 H5_REPORTS_PATH, REPORTS_PATH, PLOTS_PATH, BUNDLE_OUTPUTS, KEEP_INDIVIDUAL_OUTPUTS,
                 BIOSIMULATORS_API_ENDPOINT):
        """
        Args:
            REPORT_FORMATS (:obj:`list` of :obj:`str`): default formats to generate reports in
            PLOT_FORMATS (:obj:`list` of :obj:`str`): default formats to generate plots in
            H5_REPORTS_PATH (:obj:`str`): path to save reports in HDF5 format relative to base output directory
            REPORTS_PATH (:obj:`str`): path to save zip archive of reports relative to base output directory
            PLOTS_PATH (:obj:`str`): path to save zip archive of plots relative to base output directory
            BUNDLE_OUTPUTS (:obj:`bool`): indicates whether bundles of report and plot outputs should be produced
            KEEP_INDIVIDUAL_OUTPUTS (:obj:`bool`): indicates whether the individual output files should be kept
            BIOSIMULATORS_API_ENDPOINT (:obj:`str`): URL for BioSimulators API
        """
        self.REPORT_FORMATS = REPORT_FORMATS
        self.PLOT_FORMATS = PLOT_FORMATS
        self.H5_REPORTS_PATH = H5_REPORTS_PATH
        self.REPORTS_PATH = REPORTS_PATH
        self.PLOTS_PATH = PLOTS_PATH
        self.BUNDLE_OUTPUTS = BUNDLE_OUTPUTS
        self.KEEP_INDIVIDUAL_OUTPUTS = KEEP_INDIVIDUAL_OUTPUTS
        self.BIOSIMULATORS_API_ENDPOINT = BIOSIMULATORS_API_ENDPOINT


def get_config():
    """ Get the configuration

    Returns:
        :obj:`Config`: configuration
    """
    return Config(
        REPORT_FORMATS=[format.strip().lower() for format in os.environ.get('REPORT_FORMATS', 'csv, h5').split(',')],
        PLOT_FORMATS=[format.strip().lower() for format in os.environ.get('PLOT_FORMATS', 'pdf').split(',')],
        H5_REPORTS_PATH=os.environ.get('H5_REPORTS_PATH', 'reports.h5'),
        REPORTS_PATH=os.environ.get('REPORTS_PATH', 'reports.zip'),
        PLOTS_PATH=os.environ.get('PLOTS_PATH', 'plots.zip'),
        BUNDLE_OUTPUTS=os.environ.get('BUNDLE_OUTPUTS', '1').lower() in ['1', 'true'],
        KEEP_INDIVIDUAL_OUTPUTS=os.environ.get('KEEP_INDIVIDUAL_OUTPUTS', '1').lower() in ['1', 'true'],
        BIOSIMULATORS_API_ENDPOINT=os.environ.get('BIOSIMULATORS_API_ENDPOINT', 'https://api.biosimulators.org/'),
    )



""" Utilities for executing tasks in SED documents

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from ..config import get_config
from ..plot.data_model import PlotFormat
from ..report.data_model import DataGeneratorVariableResults, OutputResults, ReportFormat
from ..report.io import ReportWriter
from .data_model import SedDocument, Task, Report
from .io import SedmlSimulationReader
from .utils import apply_changes_to_xml_model, get_variables_for_task
import copy
import numpy
import os
import pandas
import tempfile
import types  # noqa: F401


__all__ = [
    'exec_doc',
]


def exec_doc(doc, working_dir, task_executer, base_out_path, rel_out_path=None,
             apply_xml_model_changes=False, report_formats=None, plot_formats=None):
    """ Execute the tasks specified in a SED document and generate the specified outputs

    Args:
        doc (:obj:`SedDocument` of :obj:`str`): SED document or a path to SED-ML file which defines a SED document
        working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)
        task_executer (:obj:`types.FunctionType`): function to execute each task in the SED-ML file.
            The function must implement the following interface::

                def exec_task(task, variables):
                    ''' Execute a simulation and return its results

                    Args:
                       task (:obj:`Task`): task
                       variables (:obj:`list` of :obj:`DataGeneratorVariable`): variables that should be recorded

                    Returns:
                       :obj:`DataGeneratorVariableResults`: results of variables
                    '''
                    pass

        out_path (:obj:`str`): path to store the outputs

            * CSV: directory in which to save outputs to files
              ``{out_path}/{rel_out_path}/{report.id}.csv``
            * HDF5: directory in which to save a single HDF5 file (``{out_path}/reports.h5``),
              with reports at keys ``{rel_out_path}/{report.id}`` within the HDF5 file

        rel_out_path (:obj:`str`, optional): path relative to :obj:`out_path` to store the outputs
        apply_xml_model_changes (:obj:`bool`, optional): if :obj:`True`, apply any model changes specified in the SED-ML file before
            calling :obj:`task_executer`.
        report_formats (:obj:`list` of :obj:`ReportFormat`, optional): report format (e.g., csv or h5)
        plot_formats (:obj:`list` of :obj:`PlotFormat`, optional): plot format (e.g., pdf)
    """
    # process arguments
    if not isinstance(doc, SedDocument):
        doc = SedmlSimulationReader().run(doc)
    else:
        doc = copy.deepcopy(doc)

    if report_formats is None:
        report_formats = [ReportFormat(format_value) for format_value in get_config().REPORT_FORMATS]

    if plot_formats is None:
        plot_formats = [PlotFormat(format_value) for format_value in get_config().PLOT_FORMATS]

    # apply changes to models
    modified_model_filenames = []
    for model in doc.models:
        if not os.path.isabs(model.source):
            model.source = os.path.join(working_dir, model.source)

        if apply_xml_model_changes and model.changes:
            original_model_filename = model.source

            modified_model_file, modified_model_filename = tempfile.mkstemp(suffix='.xml')
            os.close(modified_model_file)

            apply_changes_to_xml_model(model.changes, original_model_filename, modified_model_filename)

            model.source = modified_model_filename

            modified_model_filenames.append(modified_model_filename)

    # execute simulations
    variable_results = DataGeneratorVariableResults()
    for task in doc.tasks:
        if isinstance(task, Task):
            # get a list of the variables that the task needs to record
            task_vars = get_variables_for_task(doc, task)

            # execute task and record variables
            task_variable_results = task_executer(task, task_vars)

            for var in task_vars:
                variable_results[var.id] = task_variable_results.get(var.id, None)
                if variable_results[var.id] is None:
                    raise ValueError('Variable {} must be generated for task {}'.format(var.id, task.id))

        else:
            raise NotImplementedError('Tasks of type {} are not supported'.format(task.__class__.__name__))

    # generate outputs
    report_results = OutputResults()
    for output in doc.outputs:
        if isinstance(output, Report):
            dataset_ids = []
            dataset_results = []
            dataset_shapes = set()

            for data_set in output.data_sets:
                if len(data_set.data_generator.variables) != 1 or data_set.data_generator.parameters:
                    raise NotImplementedError('Data generator {} must be equal to a single variable'.format(data_set.data_generator.id))
                if (
                    len(data_set.data_generator.variables) == 1
                    and not data_set.data_generator.parameters
                    and data_set.data_generator.math != data_set.data_generator.variables[0].id
                ):
                    raise ValueError('Math of data generator must be equal to the id of the variable')
                dataset_ids.append(data_set.id)
                var_res = variable_results[data_set.data_generator.variables[0].id]
                dataset_results.append(var_res)
                dataset_shapes.add(var_res.shape)

            if len(dataset_shapes) > 1:
                raise ValueError('Data generators for report {} must have consistent shapes'.format(output.id))

            output_df = pandas.DataFrame(numpy.array(dataset_results), index=dataset_ids)
            report_results[output.id] = output_df

            for report_format in report_formats:
                ReportWriter().run(output_df,
                                   base_out_path,
                                   os.path.join(rel_out_path, output.id) if rel_out_path else output.id,
                                   format=report_format)

        else:
            raise NotImplementedError('Outputs of type {} are not supported'.format(output.__class__.__name__))

    # cleanup modified models
    for modified_model_filename in modified_model_filenames:
        os.remove(modified_model_filename)

    return (report_results, variable_results)



""" Utilities for working with SED documents

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import (SedDocument, ModelAttributeChange, Task, Report, Plot2D, Plot3D,  # noqa: F401
                         DataGeneratorVariable)
from lxml import etree
import re

__all__ = [
    'append_all_nested_children_to_doc',
    'apply_changes_to_xml_model',
]


def append_all_nested_children_to_doc(doc):
    """ Append all nested children to a SED document

    Args:
        doc (:obj:`SedDocument`): SED document
    """
    data_generators = set(doc.data_generators)
    tasks = set(doc.tasks)
    simulations = set(doc.simulations)
    models = set(doc.models)

    for output in doc.outputs:
        if isinstance(output, Report):
            for data_set in output.data_sets:
                if data_set.data_generator:
                    data_generators.add(data_set.data_generator)

        elif isinstance(output, Plot2D):
            for curve in output.curves:
                if curve.x_data_generator:
                    data_generators.add(curve.x_data_generator)
                if curve.y_data_generator:
                    data_generators.add(curve.y_data_generator)

        elif isinstance(output, Plot3D):
            for surface in output.surfaces:
                if surface.x_data_generator:
                    data_generators.add(surface.x_data_generator)
                if surface.y_data_generator:
                    data_generators.add(surface.y_data_generator)
                if surface.z_data_generator:
                    data_generators.add(surface.z_data_generator)

    for data_gen in data_generators:
        for var in data_gen.variables:
            if var.task:
                tasks.add(var.task)
            if var.model:
                models.add(var.model)

    for task in tasks:
        if isinstance(task, Task):
            if task.model:
                models.add(task.model)
            if task.simulation:
                simulations.add(task.simulation)

    doc.models += list(models - set(doc.models))
    doc.simulations += list(simulations - set(doc.simulations))
    doc.tasks += list(tasks - set(doc.tasks))
    doc.data_generators += list(data_generators - set(doc.data_generators))


def get_variables_for_task(doc, task):
    """ Get the variables that a task must record

    Args:
        doc (:obj:`SedDocument`): SED document
        task (:obj:`Task`): task

    Returns:
        :obj:`list` of :obj:`DataGeneratorVariable`: variables that task must record
    """
    variables = set()
    for data_gen in doc.data_generators:
        for var in data_gen.variables:
            if var.task == task:
                variables.add(var)
    return list(variables)


def apply_changes_to_xml_model(changes, in_model_filename, out_model_filename, pretty_print=False):
    """ Modify an XML-encoded model according to the model attribute changes in a simulation

    Args:
        changes (:obj:`list` of :obj:`ModelAttributeChange`): changes
        in_model_filename (:obj:`str`): path to model
        out_model_filename (:obj:`str`): path to save modified model
        pretty_print (:obj:`bool`, optional): if :obj:`True`, pretty print output
    """
    # read model
    et = etree.parse(in_model_filename)

    # get namespaces
    root = et.getroot()
    namespaces = root.nsmap
    if None in namespaces:
        namespaces.pop(None)
        match = re.match(r'^{(.*?)}(.*?)$', root.tag)
        if match:
            namespaces[match.group(2)] = match.group(1)

    # apply changes
    for change in changes:
        if not isinstance(change, ModelAttributeChange):
            raise NotImplementedError('Change{} of type {} is not supported'.format(
                ' ' + change.name if change.name else '', change.__class__.__name__))

        # get object to change
        obj_xpath, sep, attr = change.target.rpartition('/@')
        if sep != '/@':
            raise ValueError('target {} is not a valid XPATH to an attribute of a model element'.format(change.target))
        objs = et.xpath(obj_xpath, namespaces=namespaces)
        if len(objs) != 1:
            raise ValueError('xpath {} must match a single object in {}'.format(obj_xpath, in_model_filename))
        obj = objs[0]

        # change value
        obj.set(attr, change.new_value)

    # write model
    et.write(out_model_filename, xml_declaration=True, encoding="utf-8", standalone=False, pretty_print=pretty_print)



""" Data model for plots

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import enum


__all__ = ['PlotFormat']


class PlotFormat(str, enum.Enum):
    """ Format of a report """
    jpg = 'jpg'
    pdf = 'pdf'
    png = 'png'
    svg = 'svg'



""" Methods for submitting simulators to the BioSimulators registry

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-06
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import SimulatorSubmission  # noqa: F401
import json
import requests

GH_ISSUE_ENDPOINT = 'https://api.github.com/repos/biosimulators/Biosimulators/issues'


__all__ = ['submit_simulator_to_biosimulators_registry', 'build_gh_issue_body']


def submit_simulator_to_biosimulators_registry(simulator, gh_username, gh_access_token):
    """ Submit a version of a simulation tool for review for inclusion in the BioSimulators registry.
    This will create a GitHub issue which the BioSimulators Team will use to review your submission.

    This method requires a GitHub access and personal access token. This access token must have the `public_repo` scope.
    Instructions for creating an access token are available in the
    `GitHub documentation <https://docs.github.com/en/free-pro-team@latest/github/authenticating-to-github/creating-a-personal-access-token`>_.

    Args:
        simulator (:obj:`SimulatorSubmission`): simulator
        gh_username (:obj:`str`): GitHub username (e.g., `jonrkarr`)
        gh_access_token (:obj:`str`): GitHub personal access token.

    Raises:
        :obj:`requests.exceptions.HTTPError`: if the simulator is not successfully submitted for review
    """

    headers = {
        'Accept': 'application/vnd.github.v3+json',
    }

    data = {
        "labels": ["Validate/submit simulator"],
        "title": "Submit {} {}".format(simulator.id, simulator.version),
        "body": build_gh_issue_body(simulator),
    }

    response = requests.post(GH_ISSUE_ENDPOINT,
                             headers=headers,
                             auth=(gh_username, gh_access_token),
                             json=data)
    response.raise_for_status()


def build_gh_issue_body(submission):
    """ Build the body of a GitHub issue for the submission of a simulator

    Args:
        submission (:obj:`SimulatorSubmission`): simulator

    Returns:
        :obj:`str`: body for a GitHub issue for the submission of a simulator
    """
    return "\n".join([
        "---",
        "id: {}".format(submission.id),
        "version: {}".format(submission.version),
        "specificationsUrl: {}".format(submission.specifications_url),
        "specificationsPatch: {}".format(json.dumps(submission.specifications_patch)),
        "validateImage: {}".format(submission.validate_image),
        "commitSimulator: {}".format(submission.commit_simulator),
        "",
        "---",
    ])



""" Utilities for GitHub action workflows for reviewing and committing simulators to the BioSimulators registry

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-12-08
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

import abc
import functools
import io
import os
import requests
import types  # noqa: F401
import yamldown

__all__ = [
    'GitHubAction',
    'GitHubActionErrorHandling',
    'GitHubActionCaughtError',
]


class GitHubActionCaughtError(Exception):
    """ An error caught during the execution of a GitHub action """
    pass  # pragma: no cover


class GitHubActionErrorHandling(object):
    """ Methods for handing errors in the execution of GitHu actions """
    @classmethod
    def catch_errors(cls, uncaught_exception_msg_func=None,
                     caught_error_labels=None,
                     uncaught_error_labels=None):
        """ Generator for a decorator for CI actions that catches errors and reports them as comments to an issue

        Args:
            uncaught_exception_msg_func (:obj:`types.FunctionType`, optional): function to calculate error message to display to users
            caught_error_labels (:obj:`list` of :obj:`str`, optional): labels to apply to caught errors
            uncaught_error_labels (:obj:`list` of :obj:`str`, optional): labels to apply to uncaught errors
        """
        if uncaught_exception_msg_func is None:
            uncaught_exception_msg_func = cls.get_uncaught_exception_msg
        return functools.partial(cls._catch_errors, uncaught_exception_msg_func,
                                 caught_error_labels or [], uncaught_error_labels or [])

    @staticmethod
    def _catch_errors(uncaught_exception_msg_func, caught_error_labels, uncaught_error_labels, func):
        """ Decorator for CI actions that catches errors and reports them as comments to an issue

        Args:
            uncaught_exception_msg_func (:obj:`types.FunctionType`, optional): function to calculate error message to display to users
            caught_error_labels (:obj:`list` of :obj:`str`): labels to apply to caught errors
            uncaught_error_labels (:obj:`list` of :obj:`str`): labels to apply to uncaught errors
            func (:obj:`types.FunctionType`): decorated function
        """
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                func(*args, **kwargs)
            except GitHubActionCaughtError:
                issue_number = os.getenv('GH_ISSUE_NUMBER')
                GitHubAction.add_labels_to_issue(issue_number, caught_error_labels)
                raise
            except Exception as exception:
                issue_number = os.getenv('GH_ISSUE_NUMBER')
                GitHubAction.add_labels_to_issue(issue_number, uncaught_error_labels)
                GitHubAction.add_error_comment_to_issue(issue_number,
                                                        uncaught_exception_msg_func(exception),
                                                        raise_error=False)
                raise
        return wrapper

    def get_uncaught_exception_msg(exception):
        """ Create an error message to display to users for all exceptions not caught during the
        exception of the :obj:`run` method for a GitHub action workflow (exceptions of all types
        except :obj:`GitHubActionCaughtError`)

        Args:
            exception (:obj:`Exception`): a failure encountered during the exception of the :obj:`run`
                method for the workflow

        Returns:
            :obj:`str`: error message to display to users
        """
        return ('Sorry. We encountered an unexpected error. Our team will review the error.' +
                '\n\n  ' + str(exception).replace('\n', '\n  '))


class GitHubAction(abc.ABC):
    """ A GitHub continuous integration action

    Attributes:
        gh_auth (:obj:`tuple` of :obj:`str`): authorization for GitHub (user and access token
        gh_repo (:obj:`str`): owner and name of the repository which triggered the action
    """
    GH_ACTION_RUN_URL = 'https://github.com/{}/actions/runs/{}'
    ISSUE_ENDPOINT = 'https://api.github.com/repos/{}/issues/{}'
    ISSUE_COMMENTS_ENDPOINT = 'https://api.github.com/repos/{}/issues/{}/comments'
    ISSUE_LABELS_ENDPOINT = 'https://api.github.com/repos/{}/issues/{}/labels'
    ISSUE_ASIGNEES_ENDPOINT = 'https://api.github.com/repos/{}/issues/{}/assignees'

    def __init__(self):
        self.gh_auth = self.get_gh_auth()
        self.gh_repo = self.get_gh_repo()

    @abc.abstractmethod
    def run(self):
        pass  # pragma: no cover

    @staticmethod
    def get_gh_repo():
        """ Get the owner and name of the repository which triggered the action

        Returns:
            :obj:`str`: owner and name of the repository which triggered the action
        """
        return os.getenv('GH_REPO')

    @staticmethod
    def get_issue_number():
        """ Get the number of the issue which triggered the action

        Returns:
            :obj:`str`: issue number
        """
        return os.getenv('GH_ISSUE_NUMBER')

    @classmethod
    def get_issue(cls, issue_number):
        """ Get the properties of the GitHub issue for the submission

        Args:
            issue_number (:obj:`str`): issue number

        Returns:
            :obj:`dict`: properties of the GitHub issue for the submission
        """
        response = requests.get(
            cls.ISSUE_ENDPOINT.format(cls.get_gh_repo(), issue_number),
            auth=cls.get_gh_auth())
        response.raise_for_status()
        return response.json()

    @staticmethod
    def get_data_in_issue(issue):
        """ Get the YAML-structured data in an issue

        Args:
           issue (:obj:`dict`): properties of the GitHub issue for the submission

        Returns:
            :obj:`object`: YAML-structured data in an issue
        """
        body = io.StringIO(issue['body'].replace('\r', ''))
        data, _ = yamldown.load(body)
        return data

    @classmethod
    def get_labels_for_issue(cls, issue_number):
        """ Get the labels for an issue

        Args:
            issue_number (:obj:`str`): issue number

        Returns:
            :obj:`list` of :obj:`str`: labels
        """
        response = requests.get(
            cls.ISSUE_LABELS_ENDPOINT.format(cls.get_gh_repo(), issue_number),
            auth=cls.get_gh_auth())
        response.raise_for_status()

        labels = []
        for label in response.json():
            labels.append(label['name'])
        return labels

    @classmethod
    def add_labels_to_issue(cls, issue_number, labels):
        """ Add one or more labels to an issue

        Args:
            issue_number (:obj:`str`): issue number
            labels (:obj:`list` of :obj:`str`): labels to add to the issue
        """
        response = requests.post(
            cls.ISSUE_LABELS_ENDPOINT.format(cls.get_gh_repo(), issue_number),
            headers={'accept': 'application/vnd.github.v3+json'},
            auth=cls.get_gh_auth(),
            json={'labels': labels})
        response.raise_for_status()

    @classmethod
    def remove_label_from_issue(cls, issue_number, label):
        """ Remove a label from an issue

        Args:
            issue_number (:obj:`str`): issue number
            label (:obj:`str`): labels to add to the issue
        """
        response = requests.delete(
            cls.ISSUE_LABELS_ENDPOINT.format(cls.get_gh_repo(), issue_number) + '/' + label,
            auth=cls.get_gh_auth())
        response.raise_for_status()

    @classmethod
    def reset_issue_labels(cls, issue_number, labels_to_remove):
        """ Reset the labels for an issue

        Args:
            issue_number (:obj:`str`): issue number
            labels_to_remove (:obj:`list` of :obj:`str`): labels to remove
        """
        labels = cls.get_labels_for_issue(issue_number)
        labels_to_remove = set(labels).intersection(set(labels_to_remove))

        for label in labels_to_remove:
            cls.remove_label_from_issue(issue_number, label)

    @classmethod
    def add_comment_to_issue(cls, issue_number, comment):
        """ Post a comment to the GitHub issue

        Args:
            issue_number (:obj:`str`): issue number
            comment (:obj:`str`): comment
        """
        response = requests.post(
            cls.ISSUE_COMMENTS_ENDPOINT.format(cls.get_gh_repo(), issue_number),
            headers={'accept': 'application/vnd.github.v3+json'},
            auth=cls.get_gh_auth(),
            json={'body': comment})
        response.raise_for_status()

    @classmethod
    def add_error_comment_to_issue(cls, issue_number, comment, raise_error=True):
        """ Post an error to the GitHub issue

        Args:
            issue_number (:obj:`str`): issue number
            comment (:obj:`str`): comment
            raise_error (:obj:`bool`, optional): if :obj:`True`, raise error

        Raises:
            :obj:`ValueError`
        """
        cls.add_comment_to_issue(issue_number, ''.join([
            '```diff\n',
            '- ' + comment.rstrip().replace('\n', '\n- ') + '\n',
            '```\n',
        ]))
        if raise_error:
            raise GitHubActionCaughtError(comment)

    @classmethod
    def close_issue(cls, issue_number):
        """ Close a GitHub issue

        Args:
            issue_number (:obj:`str`): issue number
        """
        response = requests.patch(
            cls.ISSUE_ENDPOINT.format(cls.get_gh_repo(), issue_number),
            auth=cls.get_gh_auth(),
            json={'state': 'closed'})
        response.raise_for_status()

    @staticmethod
    def get_gh_action_run_id():
        """ Get the id for the current GitHub action run

        Returns:
            :obj:`str`: GitHub action run id
        """
        return os.getenv('GH_ACTION_RUN_ID')

    @classmethod
    def get_gh_action_run_url(cls):
        """ Get the URL for a GitHub action run

        Returns:
            :obj:`str`: URL for a GitHub action run
        """
        gh_repo = cls.get_gh_repo()
        gh_action_run_id = cls.get_gh_action_run_id()
        return cls.GH_ACTION_RUN_URL.format(gh_repo, gh_action_run_id)

    @staticmethod
    def get_gh_auth():
        """ Get authorization for GitHub

        Returns:
            :obj:`tuple` of :obj:`str`: authorization for GitHub (user and access token)
        """
        user = os.getenv('GH_ISSUES_USER')
        access_token = os.getenv('GH_ISSUES_ACCESS_TOKEN')
        return (user, access_token)

    @classmethod
    def assign_issue(cls, issue_number, users):
        """ Assign an issue to a list of users

        Args:
            issue_number (:obj:`str`): issue number
            users (:obj:`list` of :obj:`str`): GitHub ids of users to assign to issue
        """
        response = requests.post(
            cls.ISSUE_ASIGNEES_ENDPOINT.format(cls.get_gh_repo(), issue_number),
            auth=cls.get_gh_auth(),
            json={'assignees': users})
        response.raise_for_status()
